# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: panels.py,v 1.2 2006/01/27 07:52:40 mbaas Exp $

import wx
import cgkit.cmds
#from cgkit import getApp, pluginmanager
from cgkit import pluginmanager
import re
from . import panelicons

HORIZONTAL = 0x01
VERTICAL   = 0x02

# Exceptions:

class DuplicateNames(Exception):
    """Exception.

    Raised when there's a name clash between two nodes."""
    pass

class LayoutError(Exception):
    """Exception.

    Raises when a widget is added to a layout that already manages this
    widget."""
    pass

######################################################################

# LayoutRepository
class LayoutRepository(object):
    def __init__(self):
        self._layouts = []


# PanelWidgetRepository
class PanelWidgetRepository(object):
    """Stores all panel widgets that have been created.
    """
    
    def __init__(self):
        """Constructor."""
        self._widgets = []

    def __iter__(self):
        """Return an iterator that iterates over all widgets.
        """
        return iter(self._widgets)

    def insert(self, widget):
        """Inserts a widget into the repository.

        Nothing happens if the widget is already present.
        
        \param widget (\c PanelWidget) Widget that should be inserted into the repository
        """

        if widget not in self._widgets:
            self._widgets.append(widget)

    

# PanelWidget
class PanelWidget(object):
    """Container for a wx widget that serves as a panel widget.

    It's not allowed to have several %PanelWidget object share the same
    wx object.

    Attributes:

    - wx (\c wx \c widget)
    - name (\c str) The name of the widget (this mirrors the name of the wx widget)
    """
    def __init__(self, wx=None):
        """Constructor.

        \param wx (\c wx \c widget) The managed wx widget
        """
        self._wx = wx
        self._active = False
        self._usedby = []

    def __str__(self):
        return '<%s "%s">'%(self.__class__.__name__, self.name)

    # isActive
    def isActive(self):
        """Check if the widget is currently active.

        \return True if active
        """
        return self._active

    # activate
    def activate(self, root):
        """Activate the node.

        This method marks the widget as active.

        Derived classes may use this method to create the wx widget. In
        this case they have to use root.wx as wx parent.        

        \param root (\c Panels) The panels object where the widget is located.
        """
        self._active = True

    # deactivate
    def deactivate(self):
        """Deactivate the node.

        This method marks the widget as inactive.
        """
        self._active = False

    # isInUse
    def isInUse(self):
        """Returns whether the widget is used by any layout or not.

        \return True if the widget is used by a layout (be it active or not).
        """
        return self._usedby!=[]

    # isUsedBy
    def isUsedBy(self, layoutroot):
        """Checks if the widget is used by the given layout.

        \param layoutroot (\c ILayoutNode) The root node of the layout in question
        """
        for n in self._usedby:
            if layoutroot==n._layoutRoot():
                return True
        return False

    # acquire
    def acquire(self, panelnode):
        """Mark the widget as being used by the given node.

        \param panelnode (\c PanelNode) The panel node that controls the position and size of the widget
        """
        self._usedby.append(panelnode)

    # release
    def release(self, panelnode):
        """Mark the widget as being no longer in use by the given node.

        \param panelnode (\c PanelNode) The panel node that was controlling the position and size of the widget
        """        
        self._usedby.remove(panelnode)

    # hide
    def hide(self):
        if self._wx!=None:
            self._wx.Hide()

    # show
    def show(self):
        if self._wx!=None:
            self._wx.Show()

    def setGeom(self, x, y, w, h):
        if self._wx!=None:
            self._wx.SetDimensions(x,y,w,h)
        

    ######################################################################
    ## protected:
    
    # "wx" property...
    
    def _getWx(self):
        """Return the encapsulated wx widget.

        This method is used for retrieving the \a wx property.

        \return wxPython object.
        """
        return self._wx

    def _setWx(self, widget):
        """Set the wx widget.

        This method is used for setting the \a wx property.

        \param widget (\c wx object) wx Widget
        """
        self._wx = widget

    wx = property(_getWx, _setWx, None, "Encapsulated wx widget")

    # "name" property...
    
    def _getName(self):
        """Return the name of the widget.

        This method is used for retrieving the \a name property.

        \return Name (\c str)
        """
        if self._wx==None:
            return ""
        else:
            return str(self._wx.GetName())

    def _setName(self, name):
        """Set the name of the widget.

        This method is used for setting the \a name property.

        \param name (\c str) New name
        """
        if self._wx!=None:
            self._wx.SetName(name)

    name = property(_getName, _setName, None, "Widget name")

######################################################################
    
# Panels
class Panels(object):
    """%Panels area.

    Attributes:

    - \a layout (\c ILayoutNode): Panel layout tree
    - \a wx (\c wx widget): Corresponding wx widget
    - \a repository (\c PanelWidgetRepository) Widget repository

    Panel nodes can be accessed in one of the following ways:

    - The attribute \a mousepanel contains the panel that's currently
      underneath the mouse pointer (or None if the mouse is outside)
    - The attribute \a activepanel contains the currently active panel
      (i.e. the one that's activatable and that received the last mouse
      click)
    - Via attribute access you can address the panels by their name
    - You can iterate over all panel nodes
    

    """
        
    def __init__(self, parent, id=-1, pos=wx.DefaultPosition, size=wx.DefaultSize, style=wx.CLIP_CHILDREN):
        """Constructor.
        """
#        wx.Window.__init__(self, parent, id, pos, size, style)

        self._wx = wx.Window(parent, id, pos, size, style)

        self._repository = PanelWidgetRepository()

        self._layout = LayoutNode(root=self)

        self.mousepanel = None
        self.activepanel = None

        self._xclick = None
        self._yclick = None
        self._xsplitter = None
        self._ysplitter = None
        self._splitter_flags = None
        self._layout_node = None

        self._current_cursor = None
        self._hcrsr = wx.StockCursor(wx.CURSOR_SIZEWE)
        self._vcrsr = wx.StockCursor(wx.CURSOR_SIZENS)
        self._hvcrsr = wx.StockCursor(wx.CURSOR_SIZING)

        wx.EVT_SIZE(self._wx, self.onSize)
        wx.EVT_LEFT_DOWN(self._wx, self.onLeftDown)
        wx.EVT_LEFT_UP(self._wx, self.onLeftUp)
        wx.EVT_MOTION(self._wx, self.onMouseMove)
        wx.EVT_PAINT(self._wx, self.onPaint)

#    def __repr__(self):
#        res = "<Panellayout>"
#        return res

    def __str__(self):
        return str(self._layout)

    def updateLayout(self):
        w,h = self._wx.GetClientSizeTuple()
        self._layout.setGeom(0,0,w,h)
        self._wx.Refresh(False)

    ######################################################################
    ## protected:
        
    def __iter__(self):
        if self._layout!=None:
            return iter(self._layout.panelNodes())
        else:
            return iter([])

    def __getitem__(self, key):
        if self._layout!=None:
            return self._layout.findNodeByName(key)
        else:
            return None

    def __getattr__(self, name):
        if self._layout!=None:
            return self._layout.findNodeByName(name)
        else:
            return None
        
    def onSize(self, event):
        w,h = event.GetSize()
        self._layout.setGeom(0,0,w,h)

    def onEnterPanel(self, panel):
        self.mousepanel = panel
        self._wx.SetCursor(wx.NullCursor)
#        print "enter", panel.name

    def onLeavePanel(self, panel):
        pass
#        print "leave", panel.name

    def onClickPanel(self, panel):
        self.activepanel = panel
        self._wx.Refresh(False)

    def onLeftDown(self, event):
        x = event.GetX()
        y = event.GetY()
        flags, node = self._layout.findPanel(x,y)
        if flags!=0:
            self._splitter_flags = flags
            self._layout_node = node
            self._xclick = x
            self._yclick = y
            self._xsplitter, self._ysplitter = node.getPixelPos()
            self._wx.CaptureMouse()
            
    def onLeftUp(self, event):
        if self._xclick!=None:
            self._xclick = None
            self._yclick = None
            self._xsplitter = None
            self._ysplitter = None
            self._splitter_flags = None
            self._layout_node = None
            self._wx.ReleaseMouse()
            self._wx.SetCursor(wx.NullCursor)
#            self.Refresh(False)
        
    def onMouseMove(self, event):
        x = event.GetX()
        y = event.GetY()
#        print x,y
        if self._xclick!=None:
            dx = x-self._xclick
            dy = y-self._yclick
            if self._splitter_flags & HORIZONTAL==0:
                dx = 0
            if self._splitter_flags & VERTICAL==0:
                dy = 0
            self._layout_node.setPixelPos(self._xsplitter+dx, self._ysplitter+dy, True)
            self._wx.Refresh(False)
        else:
            flags, node = self._layout.findPanel(x,y)
            if flags==HORIZONTAL:
                self._wx.SetCursor(self._hcrsr)
            elif flags==VERTICAL:
                self._wx.SetCursor(self._vcrsr)
            elif flags==HORIZONTAL|VERTICAL:
                self._wx.SetCursor(self._hvcrsr)
            else:
                self._wx.SetCursor(wx.NullCursor)
            
    def onPaint(self, event):
        dc = wx.PaintDC(self._wx)

        dc.BeginDrawing()

        # Paint empty panels black
        dc.SetPen(wx.TRANSPARENT_PEN)
        dc.SetBrush(wx.BLACK_BRUSH)
        for n in self:
            if n.widget==None:
                dc.DrawRectangle(n._x0, n._y0, n._width, n._height)

        # Draw a border around the current panel
        activepen = wx.Pen(wx.Color(0,0,255), 3, wx.SOLID)
        dc.SetPen(activepen)
        panel = self.activepanel
        if panel!=None:
            dc.DrawRectangle(panel._x0, panel._y0, panel._width, panel._height)

        dc.EndDrawing()

    # "layout" property...
    
    def _getLayout(self):
        """Return the layout tree.

        This method is used for retrieving the \a layout property.

        \return Layout tree (\c ILayoutNode).
        """
        return self._layout

    def _setLayout(self, layout):
        """Set the layout tree.

        This method is used for setting the \a layout property.

        \param layout (\c ILayoutNode) Layout tree.
        """
        if self._layout!=None:
            self._layout.deactivate()
        self._layout = layout
        self.updateLayout()
        self._layout.activate(self)

    layout = property(_getLayout, _setLayout, None, "Panel layout tree")

    # "wx" property...
    
    def _getWx(self):
        """Return the encapsulated wx widget.

        This method is used for retrieving the \a wx property.

        \return wxPython object.
        """
        return self._wx

    wx = property(_getWx, None, None, "Encapsulated wx widget")

    # "repository" property...
    
    def _getRepository(self):
        """Return the widget repository.

        This method is used for retrieving the \a repository property.

        \return Widget repository (\c PanelWidgetRepository)
        """
        return self._repository

    repository = property(_getRepository, None, None, "Panel widget repository")
        


######################################################################
######################################################################

# ILayoutNode
class ILayoutNode(object):
    """This is the base class for a node in the layout tree.

    This class maintains the two attributes \em _parent (parent node)
    and \em _root (corresponding Panels object). The root attribute of
    every node in a tree must always point to the same %Panels object.
    If the layout is not active, then the %Panels object is None.

    The root object is set when the layout gets activated (i.e attached
    to a %Panels object). If a new node is created inside an active layout,
    the root has to be provided in the constructor.

    Attributes:

    - \b name (\c str): Node name
    """
    
    def __init__(self, root=None, name=None):
        """Constructor.

        \param root (\c Panels) Panels object which belongs to this layout or None.
        """
        self._parent = None
        self._root = root

        # Node name
        self._name = name
        

    # activate
    def activate(self, root):
        """Activate the node.

        Attaches the layout to the Panels object \a root. If the node has
        children this method has to call their %activate() method as well.
        If there are already PanelWidgets attached to this layout this
        method has to make them visible and do the layout.

        This method is called by the %Panels object whenever the layout
        is switched.

        \pre \a root must not have another layout activated.
        \pre self must be inactive
        \param root (\c Panels) The panels object to which the layout is attached.
        \see deactivate()
        """
        pass

    # deactivate
    def deactivate(self):
        """Deactivates this node.

        Detaches a layout from its root Panels object. All the widgets
        have to be hidden during deactivation.

        This method is called by the %Panels object whenever the layout
        is switched.

        \pre The layout has to be previously activated.
        \see activate()
        """
        pass

    # isActive
    def isActive(self):
        """Checks if this layout is active or not.

        \return True if the layout is active
        """
        return self._root!=None

    # setRoot
    def setRoot(self, root):
        """Set a new root.

        If the node contains children this method has to be overwritten
        and it has to set the new root on the children as well.

        This method is for internal use to propagate a new root through
        the entire tree when the layout is activated or deactivated.

        \param root (\c Panels) New root object.
        """
        self._root = root

    # setGeom
    def setGeom(self, x, y, width, height):
        """Set the position and size of the managed area.

        \param x (\c int) X position in pixel
        \param y (\c int) Y position in pixel
        \param width (\c int) Width in pixel
        \param height (\c int) Height in pixel
        """
        pass

    # applyConstraint
    def applyConstraint(self, width, height, interactive=False):
        """Check if any active constraints accept the given width and height.

        This method checks if \a width and \a height are acceptable for
        this node. If they are, the method has to return the width and
        height unmodified, otherwise an adjusted value has to be returned
        that's as close as possible to the input values.

        The parameter \a interactive specifies if the resizing was
        directly initiated by the user (i.e. a splitter was dragged).
        The usual policy is to allow resizing in this case. Otherwise
        the size is kept fixed (for example, if you only resize the entire
        application window).

        \param width (\c int)  Width to check
        \param height (\c int)  Height to check
        \param interactive (\c bool)  Flag that specifies if the resizing
               stems from a user interaction (the user drags a splitter)
        \return Tuple (width, height) with adjusted size values
        """
        pass

    # isResizable
    def isResizable(self, direction):
        """Check if the size of the node may be changed by the user.

        \param direction (\c int) A combination of HORIZONTAL and VERTICAL
        \return True if resizable.
        """
        pass

    # panelNodes
    def panelNodes(self):
        """Return a list of all panel nodes in this subtree.

        This method only has to return the panel nodes, i.e. the leafes
        of the tree. LayoutNode objects are skipped.

        This method enables the Panels object to iterate over all
        panel nodes.

        \return A list of PanelNode objects.
        """
        pass

    # findNodeByName
    def findNodeByName(self, name):
        """Returns the node in this subtree with the given name.

        \param name (\c str) Node name
        \return Node or None.
        """
        pass

    # makeNameUnique
    def makeNameUnique(self, name):
        """Makes a given name unique by appending an appropriate number.

        The argument \a name is checked if it's already used somewhere
        in the tree. If it's unused it is returned unchanged.
        Otherwise a number is appended that makes the name unique (if
        the original name aready contained a number, this number is
        increased).

        \param name (\c str) Input name
        """
        while self.findNodeByName(name)!=None:
            m = re.search("[0-9]+$", name)
            if m==None:
                name = name+"1"
            else:
                n = int(name[m.start():])
                name = name[:m.start()]+str(n+1)
        return name
            

    # findPanel
    def findPanel(self, x, y):
        return (0, self)

    # showConfigPanel
    def showConfigPanel(self):
        """Show the configuration panel.

        A LayoutNode passes this method call forward to its children
        which finally create the panel.
        """
        pass

    # hideConfigPanel
    def hideConfigPanel(self):
        """Hide the configuration panel.

        A LayoutNode passes this method call forward to its children
        which finally remove the panel.
        """
        pass

    ######################################################################
    ## protected:

    def _layoutRoot(self):
        """Return the root layout node.

        \param Root layout node (\c ILayoutNode).
        """
        p = self
        while p._parent!=None:
            p = p._parent
        return p

    # "name" property...
    
    def _getName(self):
        """Return the node name.

        This method is used for retrieving the \a name property.

        \return Name (\c str)
        """
        return self._name

    def _setName(self, name):
        """Set the node name.

        This method is used for setting the \a name property.

        \param name (\c str) New name
        """
        # Check if the new name exists already....
        prevname = self._name
        self._name = None
        if self._layoutRoot().findNodeByName(name)!=None:
            self._name = prevname
            raise DuplicateNames('There is already a node with the name "%s"'%name)

        self._name = name
        
    name = property(_getName, _setName, None, "Name")


# LayoutNodeCoords
class LayoutNodeCoords:
    """Helper class for the LayoutNode class.
    """
    def __init__(self,x0=0,x1=0,x2=0,x3=0,y0=0,y1=0,y2=0,y3=0):
        self.x0 = x0
        self.y0 = y0
        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        self.x3 = x3
        self.y3 = y3


# LayoutNode
class LayoutNode(ILayoutNode):
    """Layout node.

    This class represents a layout scheme that can manage areas
    that are laid out as depicted in the following image:
    
    \image html "panel_layout_node.png" "Layout"

    This class is also used if the region should only be split in
    horizontal or vertical direction. If the region is split horizontally
    then x1 = x2 = x3, if it's split vertically then y1 = y2 = y3.
    The following constraints are always enforced on the coordinates:
    x0 <= x1 <= x2 <= x3 and y0 <= y1 <= y2 <= y3. If a vertical splitter
    is present then x1 < x2, if a horizontal splitter is present then
    y1 < y2. All coordinates of all layout nodes are always relative to
    the widget that's associated with the layout hierarchy.

    There are two coordinates associated with each splitter. One is
    the \em logical coordinate that lies within [0,1] (0=left/top,
    1=right/bottom) and the other is the true \em pixel \em coordinate
    relative to the parent widget. In the presence of size constraints
    among the children panels/layouts the logical coordinates remain
    unchanged whereas the true pixel position changes. The pixel coordinates
    always try to represent the logical coordinates as close as possible.
    """
    def  __init__(self, x=0, y=0, width=0, height=0, root=None, name=None,
                  splittertype=HORIZONTAL|VERTICAL, x0=0.5, y0=0.5,
                  children=None):
        """Constructor.

        The arguments should be given as keyword arguments.

        \param x (\c int) Initial X position
        \param y (\c int) Initial Y position
        \param width (\c int) Initial width
        \param height (\c int) Initial height
        \param root (\c Panels) Root object
        \param name (\c str) Node name
        \param splittertype (\c int) A combination of HORIZONTAL and VERTICAL
        \param x0 (\c float) Initial logical splitter x position
        \param y0 (\c float) Initial logical splitter y position
        \param children  A tuple with children nodes (either LayoutNode or PanelNode). The number of children is determined by the splitter type.
        """
        
        ILayoutNode.__init__(self, root=root, name=name)

        # This is the region that's managed by this node
        # This coordinates are relative to the root widget
        # x1/y1 is the true pixel coordinate of the splitter 
        self._x0 = 0
        self._y0 = 0
        self._x1 = 0
        self._y1 = 0
        self._x2 = 0
        self._y2 = 0
        self._x3 = 0
        self._y3 = 0

        # Logical splitter coordinate (0-1)
        self._splitter_x = x0
        self._splitter_y = y0

        # Splitter width in pixel
        self._splitter_width = 6

        # Splitter type
        self._splitter_type = splittertype

        # Children nodes (may not be None)
        # Upper left
        self._child00 = None
        # Upper right
        self._child01 = None
        # Lower left
        self._child10 = None
        # Lower right
        self._child11 = None
        
        if children==None:
            if splittertype==HORIZONTAL|VERTICAL:
                children = (None,None,None,None)
            else:
                children = (None, None)
        self.setChildren(children)

        self.setGeom(x, y, width, height)

    def __str__(self):
        return self._toStr(self, "", 0)

    def _toStr(self, node, res, d):
        """Helper method for the __str__ operator."""

        if isinstance(node, LayoutNode):
            s="?"
            if node._splitter_type==HORIZONTAL:
                s = "horizontal"
            elif node._splitter_type==VERTICAL:
                s = "vertical"
            elif node._splitter_type==HORIZONTAL | VERTICAL:
                s = "horizontal & vertical"
            if node.name!=None:
                name = '"%s"'%node.name
            else:
                name = "<None>"
            s = 'LayoutNode %s active:%d type:%s'%(name, node.isActive(), s)
            res += d*" "+s+"\n"
            childs = node.getChildren()
            for c in childs:
                res = self._toStr(c, res, d+2)
            return res
        elif isinstance(node, PanelNode):
            return res + d*" "+str(node)+"\n"
        else:
            return res + d*" "+"???\n"

    def activate(self, root):
        self.setRoot(root)
        self._child00.activate(root)
        self._child01.activate(root)
        self._child10.activate(root)
        self._child11.activate(root)

    def deactivate(self):
        self.setRoot(None)
        self._child00.deactivate()
        self._child01.deactivate()
        self._child10.deactivate()
        self._child11.deactivate()

    def setRoot(self, root):
        ILayoutNode.setRoot(self, root)
        self._child00.setRoot(root)
        self._child01.setRoot(root)
        self._child10.setRoot(root)
        self._child11.setRoot(root)

    def getSplitterType(self):
        """Return the type of layout node.

        Returns the direction that are split by a splitter. A horizontal
        splitter splits in X direction, a vertical splitter in Y direction.

        \return A combination of HORIZONTAL and VERTICAL
        \see setSplitterType()
        """
        return self._splitter_type

    def setSplitterType(self, stype):
        """Set the layout node type.

        \param stype (\c int) A combination of HORIZONTAL and VERTICAL
        \see getSplitterType()
        """
        self._splitter_type = stype
        

    # setGeom
    def setGeom(self, x, y, width, height):
        # Initialize coordinates...
        self._x0 = int(x)
        self._y0 = int(y)
        self._x3 = int(x+width)
        self._y3 = int(y+height)
        # The following call updates the pixel position of the splitter
        self.setLogicalPos(self._splitter_x, self._splitter_y)

    def findNodeByName(self, name):
        if self._name==name:
            return self
        
        res = self._child00.findNodeByName(name)
        if res==None:
            res = self._child01.findNodeByName(name)
        if res==None:
            res = self._child10.findNodeByName(name)
        if res==None:
            res = self._child11.findNodeByName(name)
        return res

    def showConfigPanel(self):
        self._child00.showConfigPanel()
        if self._splitter_type==HORIZONTAL:
            self._child01.showConfigPanel()
        elif self._splitter_type==VERTICAL:
            self._child10.showConfigPanel()
        else:
            self._child01.showConfigPanel()
            self._child10.showConfigPanel()
            self._child11.showConfigPanel()

    def hideConfigPanel(self):
        self._child00.hideConfigPanel()
        if self._splitter_type==HORIZONTAL:
            self._child01.hideConfigPanel()
        elif self._splitter_type==VERTICAL:
            self._child10.hideConfigPanel()
        else:
            self._child01.hideConfigPanel()
            self._child10.hideConfigPanel()
            self._child11.hideConfigPanel()    
        
    def getChildren(self):
        """Return the children nodes.

        The number of children returned depends on the splitter type
        (either 2 or 4).

        \return A tuple with two or four children nodes (\c ILayoutNode).
        """
        if self._splitter_type==HORIZONTAL:
            return self._child00, self._child01
        elif self._splitter_type==VERTICAL:
            return self._child00, self._child10
        else:
            return self._child00, self._child10, self._child01, self._child11


    def setChildren(self, childs):
        """Set the children of this node.

        \a childs is a tuple of nodes which should be set as children.
        A node may also be None in which case an empty PanelNode is set.
        The children nodes must not be part of another tree (if they are
        an exception is thrown).
        
        \param childs A sequence containing the children. The number of children depends on the splitter type (2 or 4).
        """

        if self._splitter_type==HORIZONTAL:
            self.setChild(childs[0], 0)
            self.setChild(childs[1], 1)
            self.setChild(None, 2)
            self.setChild(None, 3)
        elif self._splitter_type==VERTICAL:
            self.setChild(childs[0], 0)
            self.setChild(None, 1)
            self.setChild(childs[1], 2)
            self.setChild(None, 3)
        else:
            self.setChild(childs[0], 0)
            self.setChild(childs[1], 1)
            self.setChild(childs[2], 2)
            self.setChild(childs[3], 3)


    def childIndex(self, child):
        """Return the index of the children node.

        \a child must actually be a child of self, otherwise a ValueError
        exception is thrown.

        \param child (\c ILayoutNode) Children node
        \return Index (0-3)
        """
        
        lst = [self._child00, self._child01, self._child10, self._child11]
        return lst.index(child)

    def getChild(self, idx):
        """Return a single children node.

        \param idx (\c int) Children index (0-3)
        \return Children node (\c ILayoutNode).
        """
        if idx<0 or idx>3:
            raise IndexError("Children index out of range (%d)"%idx)

        return [self._child00, self._child01, self._child10, self._child11][idx]

    def setChild(self, child, idx):
        """Replace a single children node.

        \param child (\c ILayoutNode) New children or None
        \param idx (\c int) Children index (0-3)
        \todo Pruefen, ob Child tree ok ist (keine doppelten Widgets oder Namen)
        """
        if idx<0 or idx>3:
            raise IndexError("Children index out of range (%d)"%idx)
        if child==None:
#            if self._root!=None:
#                black = wx.Window(self._root.wx, -1)
#                black.SetBackgroundColour(wx.Colour())
#            else:
#                black = None
#            child = PanelNode(widget=black)
            child = PanelNode()
        if child._parent!=None:
            raise ValueError('The panel layout node is already part of a layout tree.')
            
        if idx==0:
            self.removeChild(self._child00)
            self._child00 = child
        elif idx==1:
            self.removeChild(self._child01)
            self._child01 = child
        elif idx==2:
            self.removeChild(self._child10)
            self._child10 = child
        elif idx==3:
            self.removeChild(self._child11)
            self._child11 = child

        child._parent = self
        if self.isActive():
            child.activate(self._root)


    def removeChild(self, child):
        """Remove a child node.

        \a child may be None in which case the method returns immediately.

        \param child (\c ILayoutNode) Children which is to be removed or None
        """
        if child==None:
            return
        
        if child==self._child00:
            self._child00 = None
        elif child==self._child01:
            self._child01 = None
        elif child==self._child10:
            self._child10 = None
        elif child==self._child11:
            self._child11 = None
        else:
            raise ValueError("Layout node is not a children node.")

        child._parent = None
        if self.isActive():
            child.deactivate()

        self._fillEmptyChildren()

    def remove(self, idx=0):
        """Remove this layout node and replace it with a children node.
        """

        child = self.getChild(idx)
        self.setChildren((None,None,None,None))

        root = self._root

        parent = self._parent
        if parent!=None:
            i = parent.childIndex(self)
            parent.removeChild(self)
            parent.setChild(child, i)
        else:
            if root!=None:
                root.layout = child

        if root!=None:
            root.updateLayout()
        
        
    # isResizable
    def isResizable(self, direction):
        res = False
        if direction&HORIZONTAL:
            res |= ( (self._child00.isResizable(HORIZONTAL) and
                      self._child10.isResizable(HORIZONTAL) )
                     or
                     (self._child01.isResizable(HORIZONTAL) and
                      self._child11.isResizable(HORIZONTAL)) )
        if direction&VERTICAL:
            res |= ( (self._child00.isResizable(VERTICAL) and
                      self._child01.isResizable(VERTICAL) )
                     or
                     (self._child10.isResizable(VERTICAL) and
                      self._child11.isResizable(VERTICAL)) )

        return res

    # applyConstraint
    def applyConstraint(self, width, height, interactive=False):
        coords = LayoutNodeCoords(self._x0, 0, 0, self._x0+width,
                                  self._y0, 0, 0, self._y0+height)
        self._calcSplitterCoords(self._splitter_x, self._splitter_y,
                                 coords, interactive)
        return (coords.x3-coords.x0, coords.y3-coords.y0)

    # panelNodes
    def panelNodes(self):
        res = []
        for c in self.getChildren():
            res += c.panelNodes()
        return res

    # isInside
    def isInside(self, x, y):
        """Check if the mouse coordinate (x,y) is inside the managed region.

        The coordinate must be given relative to the root widget.

        \param x (\c int) X coordinate
        \param y (\c int) Y coordinate
        """
        return (x>=self._x0 and x<self._x3 and y>=self._y0 and y<self._y3)

    # findPanel
    def findPanel(self, x, y):
        """Find a panel by mouse position.

        The method assumes that x,y lies somehwere in the managed region.

        The return value contains a flag that has the following values:

        - 0: A panel was hit
        - 0x1: A horizontal splitter was hit
        - 0x2: A vertical splitter was hit
        - 0x3: The middle of a splitter was hit

        \return Tuple (Flag, LayoutNode)
        """
        # Left column?
        if x<self._x1:
            # Upper left panel?
            if y<self._y1:
                return self._child00.findPanel(x,y)
            # Vertical splitter?
            elif y<self._y2:
                return (VERTICAL, self)
            # Lower left panel
            else:
                return self._child10.findPanel(x,y)
        # Horizontal splitter? (or middle)
        elif x<self._x2:
            if y>=self._y1 and y<self._y2:
                return (VERTICAL | HORIZONTAL, self)
            else:
                return (HORIZONTAL, self)
        # Right column
        else:
            # Upper right panel?
            if y<self._y1:
                return self._child01.findPanel(x,y)
            # Vertical splitter?
            elif y<self._y2:
                return (VERTICAL, self)
            # Lower right panel
            else:
                return self._child11.findPanel(x,y)

    # setLogicalPos
    def setLogicalPos(self, x0=0.5, y0=0.5, interactive=False):
        """Set the logical position of the splitters.

        The pixel position is updated as well.
        
        \param x0 (\c float) Logical X position 
        \param y0 (\c float) Logical Y position 
        """

        self._splitter_x = x0
        self._splitter_y = y0

        coords = LayoutNodeCoords(self._x0, 0, 0, self._x3,
                                  self._y0, 0, 0, self._y3)
        self._calcSplitterCoords(x0, y0, coords, interactive)
        self._x1 = coords.x1
        self._x2 = coords.x2
        self._y1 = coords.y1
        self._y2 = coords.y2
        self._update()

    # getLogicalPos
    def getLogicalPos(self):
        """Return the logical position of the splitter.

        \return Tuple (x0, y0) containing the logical position.
        """
        return self._splitter_x, self._splitter_y

    # setPixelPos
    def setPixelPos(self, x, y, interactive=False):
        """Set the pixel position of the splitter.

        This method doesn't set the pixel position directly as the
        position might break some size constraints. Instead the
        position is converted to logical coordinates and setLogicalPos()
        is called instead (so the logical position is updated as well).

        \param x (\c int) X coordinate of the destination pixel position
        \param y (\c int) Y coordinate of the destination pixel position
        """
        x0, y0 = self._pixel2logical(x, y)
        self.setLogicalPos(x0, y0, interactive)

    # getPixelPos
    def getPixelPos(self):
        """Return the pixel position of the splitter.

        \param Tuple (x, y) containing the pixel position.
        """
        return self._x1, self._y1
        

    ######################################################################
    ## protected:

    def _logical2pixel(self, x0, y0):
        """Convert logical coordinates to pixel coordinates.

        \param x0 (\c float) Logical X coordinate
        \param y0 (\c float) Logical Y coordinate
        \return Tuple (x,y) with pixel coordinates (\c int, \c int)
        """
        xmax = self._x3 - self._splitter_width
        ymax = self._y3 - self._splitter_width
        return (int((1.0-x0)*self._x0 + x0*xmax),
                int((1.0-y0)*self._y0 + y0*ymax))

    def _pixel2logical(self, x, y):
        """Convert pixel coordinates to logical coordinates.

        \param x (\c int) X coordinate
        \param y (\c int) Y coordinate
        \return Tuple (x0, y0) with logical coordinates (\c float, \c float)
        """
        w = self._x3 - self._x0 - self._splitter_width
        h = self._y3 - self._y0 - self._splitter_width
        
        if w==0:
            x0 = 0.0
        else:
            x0 = float(x-self._x0)/w
            
        if h==0:
            y0 = 0.0
        else:
            y0 = float(y-self._y0)/h
            
        return x0,y0

    def _update(self):
        """Update the children areas.

        \pre x0-x3/y0-y3 contain valid values
        """
        x0,y0 = self._x0, self._y0
        x1,y1 = self._x1, self._y1
        x2,y2 = self._x2, self._y2
        x3,y3 = self._x3, self._y3
        
        self._child00.setGeom(x0, y0, x1-x0, y1-y0)
        self._child01.setGeom(x2, y0, x3-x2, y1-y0)
        self._child10.setGeom(x0, y2, x1-x0, y3-y2)
        self._child11.setGeom(x2, y2, x3-x2, y3-y2)        

    def _calcSplitterCoords(self, x0, y0, coords, interactive):
        """Calculate new splitter positions.
      
        \param x0 (\c float) Logical X coordinate of the splitter
        \param y0 (\c float) Logical Y coordinate of the splitter
        \param coords (\c LayoutNodeCoords) Coordinates container
        \param interactive (\c bool) Interactive flag
        \pre x0,x3,y0,y3 are set in coords and x0 <= x3 and y0 <= y3
        """
        
        x,y = self._logical2pixel(x0,y0)
        coords.x1 = x
        coords.x2 = x + self._splitter_width
        coords.y1 = y
        coords.y2 = y + self._splitter_width
        self._enforceSplitterConstraints(coords)
        self._enforceSizeConstraints(coords, interactive)
        return coords
        

    def _enforceSizeConstraints(self, coords, interactive):
        """Change the splitter pixel coordinates to enforce size constraints.

        This method changes pixel coordinates of the splitter so that
        the size constraints on the children will be met. If there are
        no constraints then x0-x3/y0-y3 won't be modified.
        The logical coordinates remain unchanged in any case.
        """

        # Is the left column resizable?
        if (self._child00.isResizable(HORIZONTAL) and self._child10.isResizable(HORIZONTAL)):
            # Right column has priority:
            # Determine suggested width of the right column by asking
            # child01 and child11 in both orders and taking the maximum
            w = coords.x3-coords.x2
            w1,h = self._child01.applyConstraint(w,1, interactive)
            w1,h = self._child11.applyConstraint(w1,1, interactive)
            w2,h = self._child11.applyConstraint(w,1, interactive)
            w2,h = self._child01.applyConstraint(w2,1, interactive)
            w = max(w1,w2)
            # Set the final splitter x position 
            coords.x2 = coords.x3 - w
            coords.x1 = coords.x2 - self._splitter_width
            self._enforceSplitterConstraints(coords)
        else:
            # Left column has priority:
            # Determine suggested width of the left column by asking
            # child00 and child10 in both orders and taking the maximum
            w = coords.x1-coords.x0
            w1,h = self._child00.applyConstraint(w,1, interactive)
            w1,h = self._child10.applyConstraint(w1,1, interactive)
            w2,h = self._child10.applyConstraint(w,1, interactive)
            w2,h = self._child00.applyConstraint(w2,1, interactive)
            w = max(w1,w2)
            # Set the final splitter x position 
            coords.x1 = coords.x0 + w
            coords.x2 = coords.x1 + self._splitter_width
            self._enforceSplitterConstraints(coords)

        # Is the top row resizable?
        if (self._child00.isResizable(VERTICAL) and self._child01.isResizable(VERTICAL)):
            # Bottom row has priority:
            # Determine suggested width of the bottom row by asking
            # child10 and child11 in both orders and taking the maximum
            h = coords.y3-coords.y2
            w,h1 = self._child10.applyConstraint(1,h, interactive)
            w,h1 = self._child11.applyConstraint(1,h1, interactive)
            w,h2 = self._child11.applyConstraint(1,h, interactive)
            w,h2 = self._child10.applyConstraint(1,h2, interactive)
            h = max(h1,h2)
            # Set the final splitter y position 
            coords.y2 = coords.y3 - h
            coords.y1 = coords.y2 - self._splitter_width
            self._enforceSplitterConstraints(coords)
        else:
            # Top row has priority:
            # Determine suggested height of the top row by asking
            # child00 and child01 in both orders and taking the maximum
            h = coords.y1-coords.y0
            w,h1 = self._child00.applyConstraint(1,h, interactive)
            w,h1 = self._child01.applyConstraint(1,h1, interactive)
            w,h2 = self._child01.applyConstraint(1,h, interactive)
            w,h2 = self._child00.applyConstraint(1,h2, interactive)
            h = max(h1,h2)
            # Set the final splitter y position 
            coords.y1 = coords.y0 + h
            coords.y2 = coords.y1 + self._splitter_width
            self._enforceSplitterConstraints(coords)
        

    def _enforceSplitterConstraints(self, coords):
        """Enforces the pixel coordinate constraints on both of the splitter.

        Enforces the following constraints x0 <= x1 <= x2 <= x3 and
        y0 <= y1 <= y2 <= y3 provided that x0 <= x3 and y0 <= y3 already
        hold. The distance x2-x1 resp. y2-y1 is kept intact, unless x2<x1
        or y2<y1 in which case both values will be the same. If a splitter
        is not present in a particular direction, the corresponding splitter
        coordinates will be fixed at x3 resp. y3.

        The logical coordinate remains unchanged.

        This method is called whenever the splitter position changes.

        \pre x0 <= x3 and y0 <= y3
        """

        if self._splitter_type & HORIZONTAL == 0:
            coords.x1 = coords.x2 = coords.x3
        if self._splitter_type & VERTICAL == 0:
            coords.y1 = coords.y2 = coords.y3
            
        # d is the current width of the splitter
        d = coords.x2-coords.x1
        if d<0:
            coords.x2 = coords.x1
            d = 0
            
        if coords.x1<coords.x0:
            coords.x1 = coords.x0
            coords.x2 = coords.x1 + d
        if coords.x2>coords.x3:
            coords.x2 = coords.x3
            coords.x1 = max(coords.x0, coords.x2-d)

        # d is the current width of the splitter
        d = coords.y2-coords.y1
        if d<0:
            coords.y2 = coords.y1
            d = 0
            
        if coords.y1<coords.y0:
            coords.y1 = coords.y0
            coords.y2 = coords.y1 + d
        if coords.y2>coords.y3:
            coords.y2 = coords.y3
            coords.y1 = max(coords.y0, coords.y2-d)

    def _fillEmptyChildren(self):
        if self._child00==None:
            self.setChild(PanelNode(), 0)
        if self._child01==None:
            self.setChild(PanelNode(), 1)
        if self._child10==None:
            self.setChild(PanelNode(), 2)
        if self._child11==None:
            self.setChild(PanelNode(), 3)

# PanelNode
class PanelNode(ILayoutNode):
    """This class represents a leaf node in the layout tree.

    Each node in the layout tree must be a %PanelNode object. This object
    represents one panel is associated with a wx widget. The class
    has the following attributes:

    - \a constraint (\c SizeConstraint): A constraint on the size of the panel
    - \a widget (\c PanelWidget): The associated widget
    - [\a wx (\c wx \c widget): The associated wx widget]
    """
    
    def __init__(self, name="", activatable=False, constraint=None, widget=None):
        """Constructor.

        \param name (\c str) Node name
        \param activatable (\c bool) A flag that determines if the panel can
               be activated or not (only 3D views should be made activatable)
        \param constraint (\c SizeConstraint) Size constraint
        \param widget (\c PanelWidget) The panel widget or a sequence of widgets or None
        """
        ILayoutNode.__init__(self, name=name)

        # Position and size of the panel
        self._x0 = 0
        self._y0 = 0
        self._width = 0
        self._height = 0

        # Activatable flag
        self._activatable = activatable

        # Panel widgets. This is actually a stack where the last widget
        # is on top (all other widgets are hidden).
        self._widgets = []

        # Hide all widgets
        for w in self._widgets:
            w.hide()

        # Push the widgets...
        if widget!=None:
            try:
                wlst = list(widget)
            except TypeError:
                wlst = [widget]
            
            for w in wlst:
                self.pushWidget(w)

        self.showConfigPanel()

        # Constraint
        if constraint==None:
            constraint = NoConstraint()
        self._constraint = constraint

    def __str__(self):
        if self.name!=None:
            name = '"%s"'%self.name
        else:
            name = "<None>"
        return 'PanelNode %s active:%d widgets:%s'%(name, self.isActive(), [str(x) for x in self._widgets])

    def setGeom(self, x, y, width, height):
#        print "Panel '%s', setGeom(%d, %d, %d, %d)"%(self.name, x, y, width, height)
        self._x0 = x
        self._y0 = y
        self._width = width
        self._height = height

        x+=1
        y+=1
        width-=2
        height-=2
        self._constraint.setSize(width, height)
        if len(self._widgets)>0:
            self._widgets[-1].setGeom(x,y,width,height)

    def activate(self, root):
        # Store root (Panels object)
        self.setRoot(root)

        # Mark all widgets as activated
        for w in self._widgets:
            if not isinstance(w, PanelNodeDlg):
                self._root.repository.insert(w)
            w.activate(root)

        # Set the position and size of the panel widget
        self.setGeom(self._x0, self._y0, self._width, self._height)
        # Connect and show the top widget
        if len(self._widgets)>0:
            w = self._widgets[-1]
            self._connectPanelEvents(w.wx)
            w.show()

    def deactivate(self):
        # Disconnect and hide the top widget
        if len(self._widgets)>0:
            w = self._widgets[-1]
            w.hide()
            self._disconnectPanelEvents(w.wx)
        
        # Mark all widgets as inactivated
        for w in self._widgets:
            w.deactivate()

        self.setRoot(None)


    def pushWidget(self, widget):
        """Add a widget on top of the stack.

        Adds a widget on top of the stack. If the layout is active the
        new widget will be displayed.

        \param widget (\c PanelWidget) The new panel widget
        """
        if widget.isUsedBy(self._layoutRoot()):
            raise LayoutError('The widget "%s" is already managed by this layout.'%widget.name)

        # Hide and disconnect the previous top widget...
        if self.isActive():
            if len(self._widgets)>0:
                w = self._widgets[-1]
                w.hide()
                self._disconnectPanelEvents(w.wx)

        # Store the widget...
        widget.acquire(self)
        self._widgets.append(widget)

        if self.isActive():
            if not isinstance(widget, PanelNodeDlg):
                self._root.repository.insert(widget)
            # activate the widget...
            widget.activate(self._root)
            # Set the position and size of the panel widget
            self.setGeom(self._x0, self._y0, self._width, self._height)
            # Connect and show the top widget
            if len(self._widgets)>0:
                w = self._widgets[-1]
                self._connectPanelEvents(w.wx)
                w.show()

    def popWidget(self):
        """Remove the top widget from the stack.

        \pre The stack must not be empty
        """
        w = self._widgets[-1]
        self._widgets.pop()
        w.release(self)
        # Disconnect and hide the top widget and show and connect the
        # previous one
        if self.isActive():
            w.deactivate()
            w.hide()
            self._disconnectPanelEvents(w.wx)
            if len(self._widgets)>0:
                w2 = self._widgets[-1]
                self._connectPanelEvents(w2.wx)
                w2.show()
                self.setGeom(self._x0, self._y0, self._width, self._height)
        

    def findNodeByName(self, name):
        if self._name==name:
            return self
        else:
            return None

    def showConfigPanel(self):
        # Is a config panel already on top?
        if len(self._widgets)>0 and isinstance(self._widgets[-1], PanelNodeDlg):
            return
        # Remove any config panel that's somewhere inside the stack
        self._widgets = [w for w in self._widgets if not isinstance(w, PanelNodeDlg)]
        # Create a new config panel and push it onto the stack
        dlg = PanelNodeDlg(self)
        self.pushWidget(dlg)

    def hideConfigPanel(self):
        # If there's only the config panel present then leave it there
        if len(self._widgets)==1 and isinstance(self._widgets[0], PanelNodeDlg):
            return
        
        # Is a config panel on top?
        if len(self._widgets)>0 and isinstance(self._widgets[-1], PanelNodeDlg):
            # pop the panel
            self.popWidget()
        else:
            # Remove any config panel that's somewhere inside the stack
            self._widgets = [w for w in self._widgets if not isinstance(w, PanelNodeDlg)]
            
    def isResizable(self, direction):
        return self._constraint.isResizable(direction)

    def applyConstraint(self, width, height, interactive=False):
        w,h = self._constraint(width-2, height-2, interactive)
        w+=2
        h+=2
        return w,h

    def panelNodes(self):
        return [self]

#    def findPanel(self, x, y):
#        return (0, self)

    def makeCurrent(self):
        """Make this panel the current panel.
        """
        if self._root!=None and self._activatable:
            self._root.onClickPanel(self)

    def split(self, x=None, y=None):
        """Split this panel in two or four new panels.

        \param x (\c float) Split coordinate in x direction (0-1) or None
        \param y (\c float) Split coordinate in y direction (0-1) or None
        \return Newly created layout node (\c LayoutNode) that contains
               the original panel and the new empty ones.
        """
        if x==None and y==None:
            return
        elif x!=None and y==None:
            stype = HORIZONTAL
            childs = (self, None)
        elif x==None and y!=None:
            stype = VERTICAL
            childs = (self, None)
        else:
            stype = HORIZONTAL | VERTICAL
            childs = (self, None, None, None)

        node = LayoutNode(splittertype=stype)

        parent = self._parent
        if parent!=None:
            idx = parent.childIndex(self)
            parent.removeChild(self)
            parent.setChild(node, idx)
        else:
            if self._root!=None:
                self._root.layout = node

        node.setChildren(childs)
        if self._root!=None:
            self._root.updateLayout()

        return node


    ######################################################################
    ## protected:

    def _connectPanelEvents(self, wxwin):
        """Connect the given wx window to the Panels object.

        This method sets the connections so that the Panels object will
        get notified when the mouse enters or leaves this panel and
        when it gets clicked.

        \param wxwin (\c wx \c window) A wx.Window instance
        """
#        print "connect",wxwin
        wx.EVT_ENTER_WINDOW(wxwin, self._onEnter)
        wx.EVT_LEAVE_WINDOW(wxwin, self._onLeave)
        wx.EVT_LEFT_DOWN(wxwin, self._onClick)
        wx.EVT_MIDDLE_DOWN(wxwin, self._onClick)
        wx.EVT_RIGHT_DOWN(wxwin, self._onClick)

    def _disconnectPanelEvents(self, wxwin):
        """Disconnect the given wx window from the Panels object.

        \param wxwin (\c wx \c window) A wx.Window instance
        """
#        print "disconnect",wxwin
        wx.EVT_ENTER_WINDOW(wxwin, None)
        wx.EVT_LEAVE_WINDOW(wxwin, None)
        wx.EVT_LEFT_DOWN(wxwin, None)
        wx.EVT_MIDDLE_DOWN(wxwin, None)
        wx.EVT_RIGHT_DOWN(wxwin, None)

    def _onEnter(self, event):
        self._root.onEnterPanel(self)
#        print "onEnter", self.name, self._root
        event.Skip()

    def _onLeave(self, event):
        self._root.onLeavePanel(self)
#        print "onLeave", self.name, self._root
        event.Skip()

    def _onClick(self, event):
        if self._activatable:
            self._root.onClickPanel(self)
        event.Skip()

    # "widget" property...
    
    def _getWidget(self):
        """Return the top panel widget.

        This method is used for retrieving the \a widget property.

        \return PanelWidget object or None.
        """
        if len(self._widgets)==0:
            return None
        return self._widgets[-1]

    def _setWidget(self, widget):
        """Set the top panel widget.

        This method is used for setting the \a widget property.

        \param widget (\c PanelWidget) Widget
        """
        if len(self._widgets)>0:
            self.popWidget()
        self.pushWidget(widget)

    widget = property(_getWidget, _setWidget, None, "Associated panel widget")

    # "constraint" property...
    
    def _getConstraint(self):
        """Return the constraint object.

        This method is used for retrieving the \a constraint property.

        \return Constraint object (\c SizeConstraint).
        """
        return self._constraint

    def _setConstraint(self, constraint):
        """Set the constraint object.

        This method is used for setting the \a constraint property.

        \param constraint (\c SizeConstraint) Constraint object
        """
        self._constraint = constraint

    constraint = property(_getConstraint, _setConstraint, None, "Size constraint")
    
        
        
# NoConstraint
class NoConstraint:
    """An unconstrained size constraint class."""
    
    def __call__(self, width, height, interactive=False):
        return width, height

    def setSize(self, width, height):
        pass

    def isResizable(self, direction):
        return True


# SizeConstraint
class SizeConstraint:
    """This class is used to enforce size constraints on widgets/panels.

    This class computes acceptable width/height values that conform
    to the current constraints. The constraints can either keep the
    width or height constant or at a multiple of a base width/height.
    The constraint class is a callable object that can be called
    with an input width and height and it returns a new width/height
    tuple that conforms to the constraint and is close to the input
    values.

    Examples:

    \code
    # A constraint that keeps the width fixed at 270 pixel, the height is arbitrary
    >>> c = SizeConstraint(width=270, wfixed=True)
    >>> c(100,100)
    (270, 100)

    # Now keep the height at a multiple of 32
    >>> c = SizeConstraint(height=32)
    >>> c(100,100)
    (100, 96)    
    \endcode
    
    """
    
    def __init__(self, width=None, height=None):
        """Constructor.

        \param width  A tuple (width, step, resizable) describing the width constraint
        \param height  A tuple (height, step, resizable) describing the height constraint
        """
        if width==None:
            width = None, None, False
        if height==None:
            height = None, None, False
            
        self.width, self.wstep, self.wresizable = width
        self.height, self.hstep, self.hresizable = height

    def __call__(self, width, height, interactive=False):
        """Check if the constraint accepts the given size.

        This method has to check if the given width and height
        are acceptable for the constraint. If they are the method
        just has to return those values again, if they are not
        it has to return a suggested size that would be acceptable.

        \param width (\c int) A width value
        \param height (\c int) A height value
        \param interactive (\c bool) If True the constraints also accepts multiples of its width (default: False)
        \return Tuple (width, height) specifying an acceptable size that's
               as close as possible to the input size
        """
        if self.width!=None:
            w = self.width
        else:
            w = width
            
        if self.height!=None:
            h = self.height
        else:
            h = height
        
        if self.wresizable and interactive:
            dw = width-self.width
            if dw%self.wstep==0:
                w = width
            else:
                margin = int(0.15*self.wstep)
                w = self.width + int((dw+margin)/self.wstep)*self.wstep

        if self.hresizable and interactive:
            dh = height-self.height
            if dh%self.hstep==0:
                h = height
            else:
                margin = int(0.15*self.hstep)
                h = self.height + int((dh+margin)/self.hstep)*self.hstep

        return w,h

    def setSize(self, width, height):
        """Set a new width and height.

        The new size is not set directly but run through the
        constraint (in interactive mode). So the actual size that
        gets set is a size that's compliant to the constraint.
        """
        w, h = self(width, height, True)
        if self.width!=None:
            self.width = w
        if self.height!=None:
            self.height = h

    def isResizable(self, direction):
        """Check if the size of the constraint may be changed.

        \param direction (\int) A combination of HORIZONTAL and VERTICAL
        \return True if resizable.
        """
        res = False
        if direction&HORIZONTAL:
            res |= self.wresizable
        if direction&VERTICAL:
            res |= self.hresizable
    
######################################################################


class PanelNodeDlg(PanelWidget):
    def __init__(self, panelnode):
        """Constructor.

        \param parentwin (\c MainWindow) Parent window (must have a "panels.wx" attribute)
        \param name (\c str) Widget name
        """
        PanelWidget.__init__(self)
        self.panelnode = panelnode
    
    def activate(self, root):
        PanelWidget.activate(self, root)
        self.wx = wx.Panel(root.wx, -1, style=wx.SIMPLE_BORDER)
        
        self.editlabel = wx.StaticText(self.wx, -1, "Name:")
        self.nameedit = wx.TextCtrl(self.wx, -1, value=self.panelnode.name)

        self.splitbox = wx.StaticBox(self.wx, -1, "Split")
        self.hsplitbtn = wx.BitmapButton(self.wx, -1, panelicons.getHSplitBitmap())
        self.hsplitbtn.SetToolTip(wx.ToolTip("Split this panel horizontally"))
        self.vsplitbtn = wx.BitmapButton(self.wx, -1, panelicons.getVSplitBitmap())
        self.vsplitbtn.SetToolTip(wx.ToolTip("Split this panel vertically"))
        self.hvsplitbtn = wx.BitmapButton(self.wx, -1, panelicons.getHVSplitBitmap())
        self.hvsplitbtn.SetToolTip(wx.ToolTip("Split this panel in both directions"))

        self.toplayout = wx.BoxSizer(wx.VERTICAL)

        self.s1 = wx.BoxSizer(wx.HORIZONTAL)
        self.s1.Add(self.editlabel, 0, wx.ALL | wx.ALIGN_CENTRE, 4)
        self.s1.Add(self.nameedit, 1, wx.ALL | wx.ALIGN_CENTRE, 4)

        self.s2 = wx.StaticBoxSizer(self.splitbox, wx.HORIZONTAL)
        self.s2.Add(self.hsplitbtn, 0, wx.ALL, 2)
        self.s2.Add(self.vsplitbtn, 0, wx.ALL, 2)
        self.s2.Add(self.hvsplitbtn, 0, wx.ALL, 2)

        self.toplayout.Add((0,1), 1, 0,0)
        self.toplayout.Add(self.s1, 0, wx.ALIGN_CENTRE, 0)
        self.toplayout.Add(self.s2, 0, wx.ALIGN_CENTRE, 0)
        self.toplayout.Add((0,1), 1, 0,0)

        self.wx.SetSizer(self.toplayout)

        wx.EVT_TEXT_ENTER(self.wx, self.nameedit.GetId(), self.onNameEntered)
        wx.EVT_KILL_FOCUS(self.nameedit, self.onNameEntered)
        wx.EVT_BUTTON(self.wx, self.hsplitbtn.GetId(), self.onHSplit)
        wx.EVT_BUTTON(self.wx, self.vsplitbtn.GetId(), self.onVSplit)
        wx.EVT_BUTTON(self.wx, self.hvsplitbtn.GetId(), self.onHVSplit)

    def onNameEntered(self, event):
        newname = self.nameedit.GetValue()
        try:
            self.panelnode.name = newname
        except DuplicateNames:
#            dlg = wx.MessageDialog(self.wx,
#                                   'There is already a layout node called "%s".\nPlease choose another name.'%newname, "Duplicate names", wx.OK)
#            dlg.ShowModal()
            newname = self.panelnode._layoutRoot().makeNameUnique(newname)
            self.panelnode.name = newname
            self.nameedit.SetValue(newname)
            self.nameedit.SetSelection(-1,-1)
 #           self.nameedit.SetFocus()
        
    def onHSplit(self, event):
        self.panelnode.split(x=0.5)

    def onVSplit(self, event):
        self.panelnode.split(y=0.5)

    def onHVSplit(self, event):
        self.panelnode.split(x=0.5, y=0.5)


######################################################################

class WidgetDlg(wx.Frame):
    def __init__(self, parent, id):
        wx.Frame.__init__(self, parent, id, "Panel widgets", style=wx.DEFAULT_FRAME_STYLE | wx.FRAME_FLOAT_ON_PARENT)

        self.lst = wx.ListCtrl(self, -1, style=wx.LC_REPORT | wx.LC_SINGLE_SEL | wx.LC_HRULES | wx.LC_VRULES)
        self.lst.InsertColumn(0, "Panel")
        self.lst.InsertColumn(1, "Name")
        self.lst.InsertColumn(2, "Ref.count")
        self.lst.InsertStringItem(0, "Shell")
        self.lst.SetStringItem(0,1,"shell")
        self.lst.SetStringItem(0,2,"1")
        self.lst.InsertStringItem(1, "Shader Editor")
        self.lst.SetStringItem(1,1,"shadereditor")
        self.lst.SetStringItem(1,2,"1")
        self.lst.InsertStringItem(2, "3D Viewport")
        self.lst.SetStringItem(2,1,"front")
        self.lst.SetStringItem(2,2,"1")
        self.lst.InsertStringItem(3, "3D Viewport")
        self.lst.SetStringItem(3,1,"perspective")
        self.lst.SetStringItem(3,2,"1")

def widgetDlg():
    w = WidgetDlg(getApp().window.wx, -1)
    w.Show()

cgkit.cmds.widgetDlg = widgetDlg

######################################################################

# createPanelWidget
def createPanelWidget(name, modname=None):
    """Looks for a panel widget plugin and creates an instance of it.

    \param name (\c str) Widget name
    \param modname (\c str) Module name or None (which matches every module)
    """

    for odesc in pluginmanager.iterProtoObjects(PanelWidget):
        if odesc.name==name and (modname==None or odesc.plugindesc.modulename==modname):
            ClassName = odesc.object
            widget = ClassName()
            return widget
        
    return None

cgkit.cmds.createPanelWidget = createPanelWidget

######################################################################

if __name__=="__main__":

    c = SizeConstraint(5,5, wfixed=True)
    print((c(172,22)))



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: autocam.py,v 1.1.1.1 2004/12/12 14:30:54 mbaas Exp $

## \file autocam.py
## Contains the AutoCam class.

"""This module contains the AutoCam class."""

from .Interfaces import *
from . import protocols
from . import slots
from . import _core
from .eventmanager import eventManager
from .globalscene import getScene
from . import events
from .cgtypes import *
from .component import *
       

# AutoCam
class AutoCam(Component):
    """AutoCam
    """

    protocols.advise(instancesProvide=[ISceneItem])

    def __init__(self, name="AutoCam"):
        Component.__init__(self, name=name)

        # Input
        self.input_slot = slots.Vec3Slot()
        self.addSlot("input", self.input_slot)

        # Output
        self.output_slot = slots.Vec3Slot()
        self.addSlot("output", self.output_slot)

        self.accel_factor = 0.5
        self.out_offset = vec3(0,0,0)

        self.out = vec3(0,0,0)
        self.out_velocity = vec3(0,0,0)

        eventManager().connect(events.STEP_FRAME, self)

    def protocols(self):
        return [ISceneItem, IComponent, IWorldObject, ICamera]

#    def destroy(self):
#        del self.fov_slot
#        del self.target_slot

    def onStepFrame(self):
        dt = getScene().timer().timestep
        diff = self.input-self.out
        if (diff.length()<1.0):
            a = -0.75*self.out_velocity
        else:
            a = self.accel_factor*(self.input-self.out)
        self.out_velocity += dt*a
        self.out = self.out + dt*self.out_velocity
        self.output = self.out+self.out_offset
     
    ## protected:
        
    # "input" property...
    exec(slotPropertyCode("input"))

    # "output" property...
    exec(slotPropertyCode("output"))

        




# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: glslangtokenize.py,v 1.1 2006/03/19 20:47:33 mbaas Exp $

"""OpenGL Shading Language Tokenizer."""

import re

WHITESPACE = 0
NAME       = 1
NUMBER     = 2
STRING     = 3
NEWLINE    = 4
OPERATOR   = 5
CHARACTER  = 6
TYPE       = 7
QUALIFIER  = 8

# tokenize
def tokenize(readline, tokeater):
    """Reads a Shading Language input stream and creates tokens.

    The first parameter, *readline*, must be a callable object which
    provides the same interface as the :meth:`readline()` method of built-in
    file objects. Each call to the function should return one line of
    input as a string.

    The second parameter, *tokeater*, must also be a callable object.
    It is called with six parameters: the token type, the token
    string, a tuple (*srow*, *scol*) specifying the row and column where
    the token begins in the source, a tuple (*erow*, *ecol*) giving the
    ending position of the token, the line on which the token was
    found and the filename of the current file.

    The token type can be one of

    * ``WHITESPACE``: This is a series of blanks and/or tabs.
    * ``NAME``: A valid identifier name or keyword.
    * ``NUMBER``: An integer or float.
    * ``STRING``: A string enclosed in ``'"'``.
    * ``NEWLINE``: A newline character.
    * ``OPERATOR``: An operator such as ``'+', '-', '!', '==', '!='``, etc.
    * ``CHARACTER``: A single character that doesn't fit anything else.
    * ``TYPE``: A language type (``void``, ``bool``, ``float``, ``int``, ``vec2``,
      ``vec3``, ``vec4``, etc.)
    * ``QUALIFIER``: A type qualifier (``const``, ``attribute``, ``uniform``,
      ``varying``, ``in``, ``out``, ``inout``)

    By default, the filename argument is an empty string. It will only
    be the actual filename if you provide a preprocessed file stream
    as input (so you should first run ``cpp`` on any shader). The
    tokenizer actually expects preprocessed data as it doesn't handle
    comments.
    """

    types = ["void", "bool", "float", "int", "bvec2", "bvec3", "bvec4",
             "ivec2", "ivec3", "ivec4", "vec2", "vec3", "vec4",
             "mat2", "mat3", "mat4", "struct",
             "sampler1D", "sampler2D", "sampler3D", "samplerCube",
             "sampler1DShadow", "sampler2DShadow"]
    
    qualifiers = ["const", "attribute", "uniform", "varying", "in", "out", "inout"]
                      
    regs =  ( (WHITESPACE, re.compile(r"[ \t]+")),
              (NAME,       re.compile(r"[A-Za-z_][A-Za-z_0-9]*")),
              (NUMBER,     re.compile(r"[0-9]+(\.[0-9]+)?(E(\+|-)?[0-9]+)?")),
              (STRING,     re.compile(r"\"[^\"]*\"")),
              (OPERATOR,   re.compile(r"\(|\)|\[|\]|\+\+|\-\-|\+|-|!|\.|~|\*|/|\^|%|&|\||<<|>>|<|>|<=|>=|==|!=|&&|\|\||\^\^|\?:|=|\+=|-=|\*=|/=|%=|<<=|>>=|&=|\^=|\|=")),
              (NEWLINE,    re.compile(r"\n"))
            )

    linenr   = 0
    filename = ""
    while 1:
        # Read next line
        line = readline()
        # No more lines? then finish
        if line=="":
            break

        linenr+=1
        # Base for starting column...
        scolbase = 0

        # Process preprocessor lines...
        if line[0]=="#":
            try:
                f = line.strip().split(" ")
                linenr = int(f[1])-1
                filename = f[2][1:-1]
            except:
                pass
            continue

        s = line

        # Create tokens...
        while s!="":
            unmatched=1
            # Check all regular expressions...
            for r in regs:
                m=r[1].match(s)
                # Does it match? then the token is found
                if m!=None:
                    scol = m.start()
                    ecol = m.end()
                    tok = s[scol:ecol]
                    s   = s[ecol:]
                    typ = r[0]
                    if typ==NAME:
                        if tok in types:
                            typ = TYPE
                        elif tok in qualifiers:
                            typ = QUALIFIER
                    tokeater(typ, tok, (linenr, scolbase+scol), (linenr, scolbase+ecol), line, filename)
                    scolbase += ecol
                    unmatched=0
                    continue

            # No match? then report a single character...
            if unmatched:
                tok = s[0]
                tokeater(CHARACTER, tok, (linenr, scolbase), (linenr, scolbase+1), line, filename)
                s = s[1:]
                scolbase += 1
            
            

def _tokeater(type, s, start, end, line, filename):
    if type==WHITESPACE or type==NEWLINE:
        return
    types = { 0:"WHITESPACE", 1:"NAME", 2:"NUMBER", 3:"STRING", 4:"NEWLINE",
              5:"OPERATOR", 6:"CHARACTER", 7:"TYPE", 8:"QUALIFIER" }
    print(("%s: %-10s: %-20s %s %s"%(filename, types[type], s, start, end)))

######################################################################

if __name__=="__main__":

    f=open("test.shader")
    tokenize(f.readline, _tokeater)
    



# -*- coding: utf-8 -*-

__all__ = ["kepler"]

import numpy as np

from .kepler import solve


def kepler(M, ecc, tol=1e-10):
    """Solve Kepler's equation

    Args:
        M ([type]): The mean anomaly
        ecc ([type]): The eccentricity
        tol ([type], optional): A tolerance parameter for numerical stability.
            Note: this is not an iterative solver so this is not a convergence
            parameter. You shouldn't need to change the default.
            Defaults to 1e-10.

    Raises:
        ValueError: If the eccentricity is out of bounds

    Returns:
        Three Numpy arrays with the eccentric anomaly ``E``, and the cosine and
        sine of the true anomaly ``cos(f), sin(f)``.

    """
    # Check the bounds of the eccentricity
    M, ecc = np.broadcast_arrays(M, ecc)
    if np.any(ecc < 0) or np.any(ecc >= 1.0):
        raise ValueError("ecc must be in the range 0 <= ecc < 1")

    # Solve Kepler's equation
    E = solve(M, ecc)
    cE = np.cos(E)
    sE = np.sin(E)

    # Compute cos(f), sin(f) where valid
    denom = 1 + cE
    m = denom > tol
    m_inv = ~m
    denom += 1.0 * m_inv

    # First, compute tan(0.5*E) = sin(E) / (1 + cos(E))
    tanf2 = np.sqrt((1 + ecc) / (1 - ecc)) * sE / denom  # tan(0.5*f)
    tanf2_2 = tanf2 * tanf2

    # Then we compute sin(f) and cos(f) using:
    #  sin(f) = 2*tan(0.5*f)/(1 + tan(0.5*f)^2), and
    #  cos(f) = (1 - tan(0.5*f)^2)/(1 + tan(0.5*f)^2)
    denom = 1 / (1 + tanf2_2)
    cosf = (1 - tanf2_2) * denom
    sinf = 2 * tanf2 * denom

    return E, cosf * m - 1.0 * m_inv, sinf * m



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: euleradapter.py,v 1.1.1.1 2004/12/12 14:30:59 mbaas Exp $

## \file euleradapters.py
## Contains the EulerAdapter class.

from . import protocols
from .Interfaces import *
from .component import *
from . import slots
from .cgtypes import *
from math import pi
from . import _core

# EulerAdapter
class EulerAdapter(Component):
    """Euler angle to mat3, mat4 or quat adapter.

    This class can be used to convert euler angles either to a mat3, a mat4
    or a quat. The input slots are \c anglex_slot, \c angley_slot and
    \c anglez_slot. The output slot is \c output_slot. The type of the output
    can be determined in the constructor.
    """

    protocols.advise(instancesProvide=[ISceneItem])

    def __init__(self,
                 anglex = 0,
                 angley = 0,
                 anglez = 0,
                 radians = False,
                 order = "xyz",
                 outtype = "mat3",
                 name="EulerAdapter",
                 auto_insert=True):
        """Constructor.

        \param anglex (\c float) Initial angle around x axis
        \param angley (\c float) Initial angle around y axis
        \param anglez (\c float) Initial angle around z axis
        \param radians (\c bool) True = Angles are specified in radians instead of degrees
        \param order (\c str) The rotation order ("xyz", "xzy", ...)
        \param outtpye (\c str) Output type ("mat3", "mat4", "quat")
        \param name (\c str) Component name
        \param auto_insert (\c bool) Auto insert flag
        """
        Component.__init__(self, name=name, auto_insert=auto_insert)

        if radians:
            self.factor = 1.0
        else:
            self.factor = pi/180.0

        self.anglex_slot = slots.DoubleSlot(anglex)
        self.angley_slot = slots.DoubleSlot(angley)
        self.anglez_slot = slots.DoubleSlot(anglez)
        self.addSlot("anglex", self.anglex_slot)
        self.addSlot("angley", self.angley_slot)
        self.addSlot("anglez", self.anglez_slot)

        if outtype=="mat3":
            self.output_slot = slots.ProceduralMat3Slot(self.computeMat3)
        elif outtype=="mat4":
            self.output_slot = slots.ProceduralMat4Slot(self.computeMat4)
        elif outtype=="quat":
            self.output_slot = slots.ProceduralQuatSlot(self.computeQuat)
        else:
            raise ValueError("Unknown output type: %s"%outtype)
        
        self.addSlot("output", self.output_slot)
            
        self.anglex_slot.addDependent(self.output_slot)
        self.angley_slot.addDependent(self.output_slot)
        self.anglez_slot.addDependent(self.output_slot)

        # self.fromEuler is the mat3() method that computes the matrix
        # from the euler angles. Which one exactly it is depends on the
        # order
        exec("self.fromEuler = mat3.fromEuler%s"%order.upper())

    def protocols(self):
        return [ISceneItem, IComponent]

    def computeMat3(self):
        """Slot procedure."""
        f = self.factor
        return self.fromEuler(f*self.anglex_slot.getValue(),
                              f*self.angley_slot.getValue(),
                              f*self.anglez_slot.getValue())

    def computeMat4(self):
        """Slot procedure."""
        f = self.factor
        m3 = self.fromEuler(f*self.anglex_slot.getValue(),
                            f*self.angley_slot.getValue(),
                            f*self.anglez_slot.getValue())
        res = mat4(1)
        res.setMat3(m3)
        return res

    def computeQuat(self):
        """Slot procedure."""
        f = self.factor
        m3 = self.fromEuler(f*self.anglex_slot.getValue(),
                            f*self.angley_slot.getValue(),
                            f*self.anglez_slot.getValue())
        res = quat().fromMat(m3)
        return res

    ## protected:
        
    # angle properties...
    exec(slotPropertyCode("anglex"))
    exec(slotPropertyCode("angley"))
    exec(slotPropertyCode("anglez"))

    # "output" property...
    exec(slotPropertyCode("output"))
    




# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: glrenderer.py,v 1.3 2005/08/28 19:42:43 mbaas Exp $

## \file glrenderer.py
## Contains the GLRenderer class.

from ._OpenGL.GL import *
from ._OpenGL.GLU import *
from .globalscene import getScene
from . import _core

# GLRenderInstance
class GLRenderInstance(_core.GLRenderInstance):
    """GLRenderInstance.

    """

#    protocols.advise(instancesProvide=[ISceneItem])
    
    def __init__(self):
        _core.GLRenderInstance.__init__(self)


    # pick
    def pick(self, x, y, dx=2, dy=2, root=None):
        """Do an OpenGL picking pass at the specified cursor position.

        \param x (\c float) X window coordinate where to pick
        \param y (\c float) Y window coordinate where to pick
        \param dx (\c float) Width of the picking region
        \param dy (\c float) Height of the picking region
        \param root (\c WorldObject) Only check this subtree (None=entire world).
        \return Returns a list of hits. Each hit entry is a 3-tuple (zmin, zmax, object).
        """
        
        scene = getScene()
        vpx,vpy,width,height = self.getViewport()
        if self.stereo_mode==1:
            width /= 2

        glSelectBuffer(50)
        glRenderMode(GL_SELECT)
        glInitNames()
        glPushName(0)

        glDisable(GL_DEPTH_TEST)
        glDisable(GL_LIGHTING)
        glDisable(GL_BLEND)
        glDisable(GL_NORMALIZE)

        # Projection
        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()
        gluPickMatrix(x, height-y, dx, dy, [vpx,vpy,width,height])
        glMultMatrixd(self.getProjection().toList())

        # Viewing transformation
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()
        if scene.handedness=='l':
            glScaled(-1,1,1)
        glRotated(180,0,1,0);
        glMultMatrixd(self.getViewTransformation().toList())

        # 'Render' scene
        lut = {}
        if root==None:
            root = scene.worldRoot()
        glPushMatrix()
        # Change to the coordinate system of the parent of the root
        # so that we can continue with the local system
        if root.parent!=None:
            glMultMatrixd(root.parent.worldtransform.toList())
        self._pick_node(root, 0, lut)
        glPopMatrix()

        # Get results
        buffer = glRenderMode(GL_RENDER)
        res = []
        for zmin,zmax,names in buffer:
            idx = names[0]
            res.append((zmin,zmax,lut[idx]))

        # Sort so that nearest object is first
        res.sort()
        return res

    def _pick_node(self, obj, idx, lut):
        """Helper method for the pick() method.

        Returns the current index to use for the object 'names'.
        """
        if not obj.visible:
            return idx

        glPushMatrix()
        glMultMatrixd(obj.localTransform().toList())

        geom = obj.geom
        if geom!=None:
            glLoadName(idx)
            lut[idx] = obj
            idx+=1
            geom.drawGL()

        for child in obj.iterChilds():
            idx = self._pick_node(child, idx, lut)

        glPopMatrix()
        return idx



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: mayaascii.py,v 1.10 2005/06/15 19:18:46 mbaas Exp $

import sys, types
from . import simplecpp

# The keywords that may be used for the value True
_true_keywords = ["true", "on", "yes"]
# and for the value False
_false_keywords = ["false", "off", "no"]

# splitDAGPath
def splitDAGPath(path):
    """Split a Maya DAG path into its components.

    The path is given as a string that may have the form
    <namespace>:<path> where <path> is a sequence of strings
    separated by '|'.
    The return value is a 2-tuple (namespace, names) where namespace
    is None if the path did not contain a ':'. names is a list of
    individual path names.

    Examples:

    :time1   ->  ('', ['time1'])
    |foo|bar ->  (None, ['', 'foo', 'bar'])
    foo|bar  ->  (None, ['foo', 'bar'])
    """
    if not isinstance(path, str):
        raise ValueError("string type expected as path argument, got %s"%type(path))
        
    namespace = None
    n = path.find(":")
    if n!=-1:
        namespace = path[:n]
        path = path[n+1:]
    return namespace, path.split("|")

# stripQuotes
def stripQuotes(s):
    """Remove surrounding quotes if there are any.

    The function returns the string without surrounding quotes
    (i.e. '"foo"' -> 'foo'). If there are no quotes the string
    is returned unchanged.
    """
    if s[0]=='"':
        return s[1:-1]
    else:
        return s
    

# MAPreProcessor
class MAPreProcessor(simplecpp.PreProcessor):
    """Preprocess the source file and invoke a callback for each line.
    """
    
    def __init__(self, linehandler):
        simplecpp.PreProcessor.__init__(self)
        self.linehandler = linehandler
        
    def output(self, s):
        # Ignore the preprocessor lines
        if s[0:1]!="#":
            continue_flag = self.linehandler(s)
            if not continue_flag:
                self.abort()

# PolyFace
class PolyFace:
    """Stores the data of a polyFace value.

    PolyFace objects are returned by the getValue() method of the
    Attribute class when the type was "polyFaces".
    """
    def __init__(self):
        # face (list of ints)
        self.f = None
        # holes (list of list of ints)
        # There's a list if vertex ids for each hole
        self.h = []
        # face tex coords (list of list of ints)
        self.mf = []
        # hole tex coords (list of list of ints)
        self.mh = []
        # tex coords (list of list of 2-tuple (uvset, list of ints))
        # There's a list for each loop (#f + #h).
        # Each list contains a list of 2-tuples (as there can be more than
        # one uvset per face)
        self.mu = []
        # face colors (list of list of ints)
        # There's a list for each loop (#f + #h)
        self.fc = []

    def __str__(self):
        return "<PolyFace %s #holes:%d>"%(self.f, len(self.h))

    # hasValidTexCoords
    def hasValidTexCoords(self):
        loops = [self.f]+self.h
        if len(self.mu)!=len(loops):
            return False
        for mus,loop in zip(self.mu, loops):
            if len(mus)==0:
                return False
            if len(mus[0][1])!=len(loop):
                return False

        return True

    # newLoop
    def newLoop(self):
        """This is an internal method used during construction of a poly face.

        The method has to be called after a new 'f' or 'h' attribute was
        encountered. The method will then allocate a new empty list for the
        texture coords and face colors. This list is then filled during
        construction.
        This ensures that the number of tex coords and color lists matches
        the number of loops.
        """
        self.mf.append([])
        self.mh.append([])
        self.mu.append([])
        self.fc.append([])

# NurbsCurve
class NurbsCurve:
    """Stores the data of a nurbsCurve value.

    NurbsCurve objects are returned by the getValue() method of the
    Attribute class when the type was "nurbsCurve".
    """
    def __init__(self):
        # Degree
        self.degree = 0
        # Spans
        self.spans = 0
        # Form attribute (0=open, 1=closed, 2=periodic)
        self.form = 0
        # Is the curve rational?
        self.isrational = False
        # Dimension (2 or 3)
        self.dimension = 0
        # Knots
        self.knots = []
        # Control vertices
        self.cvs = []

#    def __str__(self):
#        return "<NurbsCurve>"

# NurbsSurface
class NurbsSurface:
    """Stores the data of a nurbsSurface value.

    NurbsSurface objects are returned by the getValue() method of the
    Attribute class when the type was "nurbsSurface".
    """
    def __init__(self):
        # Degree in u and v
        self.udegree = 0
        self.vdegree = 0
        # Form attribute in u and v (0=open, 1=closed, 2=periodic)
        self.uform = 0
        self.vform = 0
        # Is the surface rational?
        self.isrational = False
        # Knots in u and v
        self.uknots = []
        self.vknots = []
        # Control vertices
        self.cvs = []

#    def __str__(self):
#        return "<NurbsSurface>"

# Attribute
class Attribute:
    """This class stores an attribute (i.e. the name and its value).

    An Attribute object is initialized with the arguments that were passed
    to the onSetAttr() callback of the reader class. The main purpose
    of an Attribute object is to convert the value into an appropriate
    Python value.
    """
    
    def __init__(self, attr, vals, opts):
        """Constructor.

        attr, vals and opts are the parameters of the onSetAttr() callback.
        """
        self._attr = stripQuotes(attr)
        self._vals = vals
        self._opts = opts

    def __str__(self):
        val = str(self._vals)
        if len(val)>10:
            val = val[:10]+"..."
        return "<Attribute %s %s %s>"%(self._attr, val, self._opts)

    # getBaseName
    def getBaseName(self):
        """Return the base name of the attribute.

        This is the first part of the attribute name (and may actually
        refer to another attribute).

        ".t"            -> "t"
        ".ed[0:11]"     -> "ed"
        ".uvst[0].uvsn" -> "uvst"
        """
        a = self._attr.split(".")
        b = a[1].split("[")
        return b[0]

    # getFullName
    def getFullName(self):
        return self._attr

    # getValue
    def getValue(self, type=None, n=None):
        """Return the converted value.

        type is a string containing the required type of the value.

        Valid types are "bool", "int", "float"
        "short2", "short3", "long2", "long3",
        "double2", "double3", "float2", "float3", "string", "int32Array",
        "doubleArray", "polyFaces", "nurbsSurface", "nurbsCurve",
        "double4", "float4" (for colors).
        """
        
        # Check if the value type was specified in the setAttr call
        valtype = self._opts.get("type", [None])[0]
        if valtype==None and type==None:
            if [x for x in self._vals if x in _true_keywords+_false_keywords]!=[]:
                valtype = "bool"
            else:
                valtype = "float"

        # Use the real attribute type if no required type was specified
        if type==None:
            type = valtype

        # Use the provided type if the attribute had no 'type' option
        if valtype==None:
            valtype = type

        # No type information given?
        if valtype==None:
            raise ValueError("No type information available for attribute '%s'"%self.getBaseName())

        # Check if the type matches the required type...
        if valtype!=type:
            raise ValueError("Attribute of type %s expected, got %s"%(type, valtype))

        # Convert the values..
        convertername = "convert"+type[0].upper()+type[1:]
        f = getattr(self, convertername)
        try:
            vs = f()
        except ValueError as e:
            sys.stderr("%s\n"%e)
            # Try a string conversion when no type was specified...
            if type==None and convertername!="convertString":
                vs = self.convertString()
            else:
                raise
        if n==None:
            return vs
        if len(vs)!=n:
            raise ValueError("%s: %d values expected, got %d"%(self._attr, n, len(vs)))
        if n==1:
            return vs[0]
        else:
            return vs

    # The following convert methods have to convert the value list (self._vals)
    # into the appropriate type. The return value is always a list of values.

    def convertInt(self):
        return [int(x) for x in self._vals]

    def convertFloat(self):
        return [float(x) for x in self._vals]

    def convertBool(self):
        res = []
        for v in self._vals:
            res.append(v in _true_keywords)
        return res

    def convertLong2(self):
        res = []
        vs = self._vals
        for i in range(0,len(vs), 2):
            res.append((int(vs[i]), int(vs[i+1])))
        return res

    convertShort2 = convertLong2

    def convertLong3(self):
        res = []
        vs = self._vals
        for i in range(0,len(vs), 3):
            res.append((int(vs[i]), int(vs[i+1]), int(vs[i+2])))
        return res

    convertShort3 = convertLong3

    def convertDouble2(self):
        res = []
        vs = self._vals
        for i in range(0,len(vs), 2):
            res.append((float(vs[i]), float(vs[i+1])))
        return res

    convertFloat2 = convertDouble2

    def convertDouble3(self):
        res = []
        vs = self._vals
        for i in range(0,len(vs), 3):
            res.append((float(vs[i]), float(vs[i+1]), float(vs[i+2])))
        return res

    convertFloat3 = convertDouble3

    def convertDouble4(self):
        res = []
        vs = self._vals
        for i in range(0,len(vs), 4):
            res.append((float(vs[i]), float(vs[i+1]), float(vs[i+2]), float(vs[i+3])))
        return res

    convertFloat4 = convertDouble4

    def convertString(self):
        return [str(x) for x in self._vals]

    def convertInt32Array(self):
        n = int(self._vals[0])
        return [int(x) for x in self._vals[1:n+1]]

    def convertDoubleArray(self):
        n = int(self._vals[0])
        return [float(x) for x in self._vals[1:n+1]]

    def convertPolyFaces(self):
        res = []
        vs = self._vals
        i=0
        pf = None
        while i<len(vs):
            c = vs[i]
            i+=1
            if c not in ["f", "h", "mf", "mh", "mu", "fc"]:
                raise ValueError("Unknown polyFace data: %s"%c)
            
            if c=="mu":
                uvset = int(vs[i])
                i+=1
                
            n = int(vs[i])
            ids = [int(x) for x in vs[i+1:i+n+1]]
            i+=n+1
            
            # Is that already a new polyFace? Then store the previous one
            # and start a new face
            if c=="f":
                if pf!=None:
                    res.append(pf)
                pf = PolyFace()
                pf.f = ids
                pf.newLoop()
            elif c=="h":
                pf.h.append(ids)
                pf.newLoop()
            elif c=="mu":
                pf.mu[-1].append((uvset, ids))
            else:
                # Set the IDs into the corresponding list
                lst = getattr(pf, c)
                lst[-1] = ids

        # Append the last face
        if pf!=None:
            res.append(pf)
            
        return res

    def convertNurbsCurve(self):
        res = []
        vs = self._vals

        while vs!=[]:
            nc = NurbsCurve()
            nc.degree = int(vs[0])
            nc.spans = int(vs[1])
            nc.form = int(vs[2])
            nc.isrational = vs[3] in _true_keywords
            nc.dimension = int(vs[4])
            # Get knots
            n = int(vs[5])
            nc.knots = [float(x) for x in vs[6:n+6]]
            vs = vs[n+6:]
            # Get CVs
            n = int(vs[0])
            dim = nc.dimension
            if nc.isrational:
                dim += 1
            cvs = []
            for i in range(n):
                cvs.append( tuple([float(x) for x in vs[1+i*dim:1+(i+1)*dim]]) )
            nc.cvs = cvs
            vs = vs[1+n*dim:]
            res.append(nc)

        return res

    def convertNurbsSurface(self):
        res = []
        vs = self._vals

        while vs!=[]:
            ns = NurbsSurface()
            ns.udegree = int(vs[0])
            ns.vdegree = int(vs[1])
            ns.uform = int(vs[2])
            ns.vform = int(vs[3])
            ns.isrational = vs[4] in _true_keywords
            vs = vs[5:]
            # Get u knots
            n = int(vs[0])
            ns.uknots = [float(x) for x in vs[1:n+1]]
            vs = vs[n+1:]
            # Get v knots
            n = int(vs[0])
            ns.vknots = [float(x) for x in vs[1:n+1]]
            vs = vs[n+1:]
            # Skip TRIM|NOTRIM
            if vs[0] in ["TRIM", "NOTRIM"]:
                vs = vs[1:]
            # Get CVs
            n = int(vs[0])
            if ns.isrational:
                dim = 4
            else:
                dim = 3
            cvs = []
            for i in range(n):
                cvs.append( tuple([float(x) for x in vs[1+i*dim:1+(i+1)*dim]]) )
            ns.cvs = cvs
            vs = vs[1+n*dim:]
            res.append(ns)

        return res
        
    # setValue
    def setValue(self, value):
        pass

# MultiAttrStorage
class MultiAttrStorage:
    """This helper class serves as MEL style array.

    You can assign values to arbitrary indices which will automatically
    enlarge the array (filling missing values with None).
    The slicing operation is different than Python as the stop index is
    inclusive. The actual array is stored in the _array attribute.

    Example:

    a = MultiAttrStorage()
    a[2] = 5               # -> [None, None, 5]
    a[4:6] = [1,2,3]       # -> [None, None, 5, None, 1, 2, 3]


    Reading an arbitrary attribute from an object of this class will
    automatically create this attribute which will in turn be of type
    MultiAttrStorage. So this class can also be used for compound objects.
    If you want to check such a compound for a particular attribute you
    must not use hasattr() as this would silently create the attribute if
    it didn't already exist and always return True. Instead you have to
    do the check as follows:

    if <attrname> in dir(compound):
       ...
    """
    def __init__(self):
        self._array = []

    def __str__(self):
        namedattrs = [x for x in list(self.__dict__.keys()) if x[0:1]!="_"]
        # Not an array but a 'struct' with named attributes...
        if namedattrs!=[] and self._array==[]:
            a = []
            for name in namedattrs:
                a.append(".%s:%s"%(name, getattr(self, name)))
            return ", ".join(a)
        else:
            return "["+", ".join([str(x) for x in self._array])+"]"

    def __iter__(self):
        return iter(self._array)

    def __len__(self):
        return len(self._array)

    def __getattr__(self, name):
        if name[:2]=="__":
            raise AttributeError(name)
        ma = MultiAttrStorage()
        setattr(self, name, ma)
        return ma

    def __getitem__(self, key):
        if key>=len(self._array):
            ma = MultiAttrStorage()
            self.__setitem__(key, ma)
            return ma
        else:
            return self._array[key]

    def __setitem__(self, key, value):
        al = len(self._array)
        if type(key)==slice:
            if key.stop>=al:
                self._array += (key.stop-al+1)*[None]
            if len(value)!=(key.stop-key.start+1):
                raise ValueError("%d values expected, got %d"%(key.stop-key.start+1, len(value)))
            self._array[key.start:key.stop+1] = value
#            print key.start,key.stop
        else:
            if key<al:
                self._array[key] = value
            else:
                self._array += (key-al)*[None] + [value]

# Node
class Node:
    """A generic Maya node class.

    This is a helper class which may be used in a concrete implementation
    of the MAReader class to represent Maya nodes.

    This class does not implement the actual functionality of a
    particular Maya node, it just tracks attribute changes and
    connections which can later be retrieved once the entire file was
    read. So this class can be used for all Maya nodes in a file. 
    """
    
    def __init__(self, nodetype, opts, parent=None):
        """Constructor.

        nodetype and opts are the arguments of the onCreateNode()
        callback of the MAReader class.
        parent is the parent Node object or None.
        """

        # Do some type checking...
        if not isinstance(nodetype, str):
            raise ValueError("Argument 'nodetype' must be a string, got %s."%(type(nodetype)))
        if type(opts)!=dict:
            raise ValueError("Argument 'opts' must be a dict, got %s."%(type(opts)))
        if parent!=None and not isinstance(parent, Node):
            raise ValueError("Argument 'parent' must be a Node object or None, got %s."%(type(parent)))

        # A string containing the node type
        self.nodetype = nodetype
        # The options dictionary
        self.opts = opts

        # Parent Node object
        self._parent = None

        # Children Node objects
        self._children = []
        
        # Attribute values.
        # Key: Attribute base name / Value: List of Attribute objects
        self._setattr = {}

        # If True, accessing attributes will automatically create a new
        # attribute that contains a MultiAttrStorage object
        self._create_attributes = False

        # Key: Local attribute / Value: (nodename, attrname) of source
        self.in_connections = {}
        # Key: Local attribute / Value: List of (node, nodename, attrname) tuples
        self.out_connections = {}

        # Set the parent
        self.setParent(parent)

    def __str__(self):
        return '<Node "%s" %s>'%(self.getFullName(), self.nodetype)

    def __getattr__(self, name):
        if self._create_attributes:
            ma = MultiAttrStorage()
            setattr(self, name, ma)
            return ma
        raise AttributeError(name)

    # getName
    def getName(self):
        """Return the node name.

        If no node name was specified during the creation of the object,
        the dummy name 'MayaNode' is returned.
        """
        return self.opts.get("name", ["MayaNode"])[0]

    # getFullName
    def getFullName(self):
        """Return the full node name.
        """
        name = self.getName()
        if self._parent==None:
            return "|%s"%name
        else:
            return "%s|%s"%(self._parent.getFullName(), name)

    # getParentName
    def getParentName(self):
        """Return the parent's node name or None.
        """
        if self._parent==None:
            return None
        else:
            return self._parent.getFullName()

    # getParent
    def getParent(self):
        """Return the parent Node object or None.
        """
        return self._parent

    # setParent
    def setParent(self, parent):
        """Reparent the Node object.
        """
        # Remove self from the previous parent's children list...
        if self._parent!=None:
            self._parent._children.remove(self)

        # Set the new parent...
        self._parent = parent
        if parent!=None:
            parent._children.append(self)

    # iterChildren
    def iterChildren(self):
        """Return an iterator that yields all children Node objects.
        """
        return iter(self._children)

    # setAttr
    def setAttr(self, attr, vals, opts):
        """Store the attribute value.

        The arguments are the same than the arguments of the onSetAttr()
        callback in the MAReader class.
        The final Python value can be retrieved with the getAttrValue()
        method.
        """
        # Is this setattr call only used to "declare" the size of an array?
        # then ignore the call
        if vals==[] and "size" in opts:
            return
        
        a = Attribute(attr, vals, opts)
        basename = a.getBaseName()
        if basename in self._setattr:
            self._setattr[basename].append(a)
        else:
            self._setattr[basename] = [a]

    # addAttr
    def addAttr(self, opts):
        pass

    # addInConnection
    def addInConnection(self, localattr, nodename, attrname):
        """Add an 'in' connection.

        'in' = node.attr is connected to localattr
        nodename is the name of a Node object and attrname the full
        attribute name.
        """
        self.in_connections[localattr] = (nodename, attrname)

    # addutConnection
    def addOutConnection(self, localattr, node, nodename, attrname):
        """Add an 'out' connection.

        'out' = localattr is connected to node.attr
        node is a Node object, nodename the name of the node and attrname
        the full attribute name.
        """
        if not isinstance(node, Node):
            raise ValueError("Argument 'node' must be a Node instance.")
        if localattr in self.out_connections:
            self.out_connections[localattr].append((node, nodename, attrname))
        else:
            self.out_connections[localattr] = [(node, nodename, attrname)]
        

    # getAttrValue
    def getAttrValue(self, lname, sname, type, n=1, default=None):
        """Get the Python value of an attribute.

        lname is the long name, sname the short name. type is the
        required type and n the required number of elements. type and
        n may be None.
        The return value is either a normal Python type (int, float, sequence)
        or a MultiAttrStorage object in cases where the setAttr command
        contained the index operator.
        When no attribute with the given long or short name could be
        found the provided default value is returned.

        This method will only return a value other than the default
        value when the \method{setAttr()} was called with the corresponding
        attribute.
        """
        # 'Execute' the stored setAttr commands for the long name and
        # the short name. This will store the value as an attribute
        # of self.
        self._executeAttrs(lname, type, n)
        self._executeAttrs(sname, type, n)
        # Check if the long name is available, otherwise try the short name...
        if hasattr(self, lname):
            return getattr(self, lname)
        return getattr(self, sname, default)

    # getInNode
    def getInNode(self, localattr_long, localattr_short):
        """Return the node and attribute that serves as input for localattr.

        The return value is a 2-tuple (nodename, attrname) that specifies
        the input connection for localattr. (None, None) is returned if there
        is no connection.
        """
        node, attr = self.in_connections.get(localattr_long, (None, None))
        if node==None:
            node, attr = self.in_connections.get(localattr_short, (None, None))
        return node, attr

    # getOutNodes
    def getOutNodes(self, localattr_long, localattr_short):
        """Return the nodes and attributes that this attribute connects to.

        The return value is a list of 3-tuples (node, nodename, attrname) that
        specify the output connections for localattr.
        """
        lst = self.out_connections.get(localattr_long, None)
        if lst==None:
            lst = self.out_connections.get(localattr_short, [])
        return lst

    # getOutAttr
    def getOutAttr(self, localattr_long, localattr_short, dstnodetype):
        """Check if a local attribute is connected to a particular type of node.

        Returns a tuple (node, attrname) where node is the Node object
        of the destination node and attrname the name of the destination
        attribute. If there is no connection with a node of type dstnodetype,
        the method returns (None,None).
        If the attribute is connected to more than one node with the
        given type or to several attributes of the same node then only
        the first connection encountered is returned.
        """
        cs = self.getOutNodes(localattr_long, localattr_short)

        # Search the connections for one that goes to a node with
        # the specified type...
        for node,nodename,attrname in cs:
            if node.nodetype==dstnodetype:
                return node, attrname
        return None,None

    # _executeAttrs
    def _executeAttrs(self, basename, type, n):
        """Retrieve the Python values out of an attribute.

        All stored attributes with the given basename will be 'executed'.
        This means, their Python value will be obtained and stored
        in the node under the same name than the original attribute.
        type is the required type of the attribute (may be None) and
        n the required number of elements (or None = Dynamic array).
        """
        self._create_attributes = True
        efftype = type
        for attr in self._setattr.get(basename, []):
            # Keep the type information once an attribute explicitly
            # provided some info. Later attributes that don't specify
            # the type (as can happen with the texture coordinates) can
            # then be read properly.
            if type==None:
                valtype = attr._opts.get("type", [None])[0]
                if valtype!=None:
                    efftype = valtype
            v = attr.getValue(efftype, n)
            # Unpack the value if the number of elements wasn't known and
            # it turned out to be only one element (in this case, the below
            # assignment will be of the form a[i]=v instead of a[i:j]=v)
            if n==None and len(v)==1:
                v = v[0]
            if attr.getBaseName() in ["in", "from", "for", "if", "while",
                                      "else", "elif", "exec"]:
                cmd = "setattr(self, '%s', v)"%attr.getFullName()[1:]
            else:
                cmd = "self%s = v"%attr.getFullName()
            if v!=[]:
                exec(cmd)
        self._create_attributes = False

######################################################################

# MAReader
class MAReader:
    """Low level MA (Maya ASCII) reader.

    The MAReader class reads Maya ASCII files and calls handler
    methods which have to be implemented in a derived class. The
    content of the file is actually a subset of the Maya Embedded
    Language (MEL) which is the scripting language implemented inside
    Maya. The MAReader parses the file, breaks down the content of the
    file in commands and their arguments and options (expressions are
    not evaluated). Each MEL command will then trigger a callback
    method that has to execute the command. These callback methods
    have to be implemented in a derived class.

    There are 13 MEL commands that can appear in a Maya ASCII file: 

    - file 
    - requires 
    - fileInfo 
    - currentUnit 
    - createNode 
    - setAttr 
    - addAttr 
    - connectAttr 
    - disconnectAttr 
    - parent 
    - select
    - lockNode
    - relationship

    Each command has a number of arguments and can also take
    options. The callback methods receive the arguments as regular
    arguments to the method and the options as an additional argument
    opts which is a dictionary containing the options that were
    specified in the file. The key is the long name of the option
    (without leading dash) and the value is a list of strings
    containing the option values. The number of values and how they
    have to be interpreted depend on the actual option.
    
    The callbacks may access a few instance variables that carry
    further information:
    
    - filename: The name of the ma file
    - cmd_start_linenr: The line number where the current command began
    - cmd_end_linenr: The line number where the current command ended 
    """
    
    def __init__(self):

        # createNode options
        self.createNode_name_dict = { "n":"name", "p":"parent", "s":"shared", "ss":"skipSelect" }
        self.createNode_opt_def = { "name" : (1, None),
                                    "parent" : (1, None),
                                    "shared" : (0, None),
                                    "skipSelect" : (0, None) }

        # setAttr options
        self.setAttr_name_dict = { "k":"keyable",
                                   "l":"lock",
                                   "cb":"channelBox",
                                   "ca":"caching",
                                   "s":"size",
                                   "typ":"type",
                                   "av":"alteredValue",
                                   "c":"clamp" }
        self.setAttr_opt_def = { "keyable" : (1, None),
                                 "lock" : (1, None),
                                 "channelBox" : (1, None),
                                 "caching" : (1, None),
                                 "size" : (1, None),
                                 "type" : (1, None),
                                 "alteredValue" : (0, None),
                                 "clamp" : (0, None)}

        # fileInfo options
        self.fileInfo_name_dict = { "rm":"remove" }
        self.fileInfo_opt_def = { "remove" : (1, None)}

        # currentUnit options
        self.currentUnit_name_dict = { "l":"linear",
                                       "a":"angle",
                                       "t":"time",
                                       "f":"fullName",
                                       "ua":"updateAnimation"}
        self.currentUnit_opt_def = { "linear" : (1, None),
                                     "angle" : (1, None),
                                     "time" : (1, None),
                                     "fullName" : (0, None),
                                     "updateAnimation" : (1, None)}

        # connectAttr options
        self.connectAttr_name_dict = { "l":"lock",
                                       "f":"force",
                                       "na":"nextAvailable",
                                       "rd":"referenceDest" }
        self.connectAttr_opt_def = { "lock" : (1, None),
                                     "force" : (0, None),
                                     "nextAvailable" : (0, None),
                                     "referenceDest" : (1, None) }

        # disconnectAttr options
        self.disconnectAttr_name_dict = { "na":"nextAvailable" }
        self.disconnectAttr_opt_def = { "nextAvailable" : (0, None)}

        # parent options
        self.parent_name_dict = { "w":"world",
                                  "r":"relative",
                                  "a":"absolute",
                                  "add":"addObject",
                                  "rm":"removeObject",
                                  "s":"shape",
                                  "nc":"noConnections" }
        self.parent_opt_def = { "world" : (0, None),
                                "relative" : (0, None),
                                "absolute" : (0, None),
                                "addObject" : (0, None),
                                "removeObject" : (0, None),
                                "shape" : (0, None),
                                "noConnections" : (0, None) }

        # select options
        self.select_name_dict = { "adn":"allDependencyNodes",
                                  "ado":"allDagObjects",
                                  "vis":"visible",
                                  "hi":"hierarchy",
                                  "af":"addFirst",
                                  "r":"replace",
                                  "d":"deselect",
                                  "tgl":"toggle",
                                  "cl":"clear",
                                  "ne":"noExpand" }
        self.select_opt_def = { "all" : (0, None),
                                "allDependencyNodes" : (0, None),
                                "allDagObjects" : (0, None),
                                "visible" : (0, None),
                                "hierarchy" : (0, None),
                                "add" : (0, None),
                                "addFirst" : (0, None),
                                "replace" : (0, None),
                                "deselect" : (0, None),
                                "toggle" : (0, None),
                                "clear" : (0, None),
                                "noExpand" : (0, None) }

        # addAttr options
        self.addAttr_name_dict = { "ln":"longName",
                                   "sn":"shortName",
                                   "bt":"binaryTag",
                                   "at":"attributeType",
                                   "dt":"dataType",
                                   "dv":"defaultValue",
                                   "m":"multi",
                                   "im":"indexMatters",
                                   "min":"minValue",
                                   "hnv":"hasMinValue",
                                   "max":"maxValue",
                                   "hxv":"hasMaxValue",
                                   "ci":"cachedInternally",
                                   "is":"internalSet",
                                   "p":"parent",
                                   "nc":"numberOfChildren",
                                   "uac":"usedAsColor",
                                   "h":"hidden",
                                   "r":"readable",
                                   "w":"writable",
                                   "s":"storable",
                                   "k":"keyable",
                                   "smn":"softMinValue",
                                   "hsn":"hasSoftMinValue",
                                   "smx":"softMaxValue",
                                   "hsx":"hasSoftMaxValue",
                                   "en":"enumName" }
        self.addAttr_opt_def = { "longName" : (1, None),
                                 "shortName" : (1, None),
                                 "binaryTag" : (1, None),
                                 "attributeType" : (1, None),
                                 "dataType" : (1, None),
                                 "defaultValue" : (1, None),
                                 "multi" : (0, None),
                                 "indexMatters" : (1, None),
                                 "minValue" : (1, None),
                                 "hasMinValue" : (1, None),
                                 "maxValue" : (1, None),
                                 "hasMaxValue" : (1, None),
                                 "cachedInternally" : (1, None),
                                 "internalSet" : (1, None),
                                 "parent" : (1, None),
                                 "numberOfChildren" : (1, None),
                                 "usedAsColor" : (0, None),
                                 "hidden" : (1, None),
                                 "readable" : (1, None),
                                 "writable" : (1, None),
                                 "storable" : (1, None),
                                 "keyable" : (1, None),
                                 "softMinValue" : (1, None),
                                 "hasSoftMinValue" : (1, None),
                                 "softMaxValue" : (1, None),
                                 "hasSoftMaxValue" : (1, None),
                                 "enumName" : (1, None) }

        # file options (incomplete)
        self.file_name_dict = { "bls":"buildLoadSettings",
                                "c":"command",
                                "dns":"defaultNamespace",
                                "dr":"deferReference",
                                "f":"force",
                                "fr":"flushReference",
                                "gl":"groupLocator",
                                "gn":"groupName",
                                "gr":"groupReference",
                                "ir":"importReference",
                                "lck":"lockReference",
                                "lf":"lockFile",
                                "lrd":"loadReferenceDepth",
                                "lad":"loadAllDeferred",
                                "lar":"loadAllReferences",
                                "lnr":"loadNoReferences",
                                "lr":"loadReference",
                                "ls":"loadSettings",
                                "pr":"preserveReferences",
                                "new":"newFile",
                                "o":"open",
                                "op":"options",
                                "pmt":"prompt",
                                "r":"reference",
                                "ra":"renameAll",
                                "rdi":"referenceDepthInfo",
                                "rfn":"referenceNode",
                                "rpr":"renamingPrefix",
                                "shd":"sharedNodes",
                                "sns":"swapNamespace",
                                "srf":"sharedReferenceFile",
                                "str":"strict",
                                "ns":"namespace",
                                # The following flags are not documented in the Maya docs
                                "pm":"proxyManager",
                                "pt":"proxyTag",
                                "ap":"activeProxy" }
        
        self.file_opt_def = { "buildLoadSettings" : (0, None), 
                              "command" : (1, None),
                              "defaultNamespace" : (0, None),
                              "deferReference" : (1, None),
                              "force" : (0, None),
                              "flushReference" : (1, None),
                              "groupLocator" : (0, None),
                              "groupName" : (1, None),
                              "groupReference" : (0, None),
                              "importReference" : (0, None),
                              "lockReference" : (0, None),
                              "lockFile" : (1, None),
                              "loadReferenceDepth" : (1, None),
                              "loadAllDeferred" : (1, None),
                              "loadAllReferences" : (0, None),
                              "loadNoReferences" : (0, None),
                              "loadReference" : (1, None),
                              "loadSettings" : (1, None),
                              "preserveReferences" : (0, None),
                              "newFile" : (0, None),
                              "open" : (0, None),
                              "options" : (1, None),
                              "prompt" : (1, None),
                              "reference" : (0, None),
                              "renameAll" : (1, None),
                              "referenceDepthInfo" : (1, None),
                              "referenceNode" : (1, None),
                              "renamingPrefix" : (1, None),
                              "sharedNodes" : (1, None),
                              "swapNamespace" : (2, None),
                              "sharedReferenceFile" : (0, None),
                              "strict" : (1, None),
                              "namespace" : (1, None),
                              "proxyManager" : (1, None),
                              "proxyTag" : (1, None),
                              "activeProxy" : (0, None) }

        # lockNode options
        self.lockNode_name_dict = { "l":"lock",
                                    "ic":"ignoreComponents" }
        self.lockNode_opt_def = { "lock" : (1, None),
                                  "ignoreComponents" : (0, None) }

        # relationship options
        self.relationship_name_dict = { "b":"break" }
        self.relationship_opt_def = { "break" : (0, None)}

        
    # Provide linenr as an alias for cmd_start_linenr
    @property
    def linenr(self):
        return self.cmd_start_linenr
    
    def read(self, f):
        """Read a MA file and invoke the callbacks.

        f is a file-like object or the name of a file.
        """
        self.begin()
        if isinstance(f, str):
            self.filename = f
        else:
            self.filename = getattr(f, "name", "?")
        # A flag that indicates if a new MEL command is about to begin
        self.new_cmd = True
        # The name of the current MEL command
        self.cmd = None
        # The arguments of the current MEL command
        self.args = None
        # This flag specifies whether reading the file should continue or not
        self.continue_flag = True

        # The line number where the current MEL command began
        self.cmd_start_linenr = None
        # The line number where the current MEL command ended
        self.cmd_end_linenr = None
        
        cpp = MAPreProcessor(self.lineHandler)
        self.cpp = cpp
        # Read the file and invoke the lineHandler for each line...
        cpp(f)
       
        # Execute the last command
        self.processCommands(";")
        self.end()

    def lineHandler(self, s):
#        self.linenr += 1
        z = s.strip()
        if z!="":
            self.processCommands(z)
        return self.continue_flag
    
    def abort(self):
        """Stop reading the MA file.
        
        This method can be called by a callback method to abort
        reading the file.
        """
        self.continue_flag = False

    def begin(self):
        """Callback that is invoked before the file is read."""
        pass

    def end(self):
        """Callback that is invoked after the file was read."""
        pass

    def onFile(self, filename, opts):
        """Callback for the 'file' MEL command."""
        pass
#        print "file", filename, opts

    def onRequires(self, product, version):
        """Callback for the 'requires' MEL command."""
        pass
#        print "requires",product, version

    def onFileInfo(self, keyword, value, opts):
        """Callback for the 'fileInfo' MEL command."""
        pass
#        print "fileInfo",keyword, value

    def onCurrentUnit(self, opts):
        """Callback for the 'currentUnit' MEL command."""
        pass
#        print "currentUnit",opts

    def onCreateNode(self, nodetype, opts):
        """Callback for the 'createNode' MEL command."""
        pass
#        print "createNode", nodetype, opts

    def onSetAttr(self, attr, vals, opts):
        """Callback for the 'setAttr' MEL command."""
        pass
#        print "setAttr %s = %s %s"%(attr, vals, opts)

    def onConnectAttr(self, srcattr, dstattr, opts):
        """Callback for the 'connectAttr' MEL command."""
        pass
#        print "connectAttr %s %s %s"%(srcattr, dstattr, opts)

    def onDisconnectAttr(self, srcattr, dstattr, opts):
        """Callback for the 'disconnectAttr' MEL command."""
        pass
#        print "disconnectAttr %s %s %s"%(srcattr, dstattr, opts)

    def onAddAttr(self, opts):
        """Callback for the 'addAttr' MEL command."""
        pass
#        print "addAttr", opts

    def onParent(self, objects, parent, opts):
        """Callback for the 'parent' MEL command."""
        pass
#        print "parent",objects,parent,opts

    def onSelect(self, objects, opts):
        """Callback for the 'select' MEL command."""
        pass
#        print "select",objects,opts

    def onLockNode(self, objects, opts):
        """Callback for the 'lockNode' MEL command.
        
        objects is a list of objects (which may be empty).
        """
        pass
#        print "lockNode",objects,opts

    def onRelationship(self, args, opts):
        """Callback for the 'relationship' MEL command.
        
        args is a list of the command line arguments. opts is the options dict.
        """
        pass
#        print "relationship",args,opts

    # onCommand
    def onCommand(self, cmd, args):
        """Generic command callback.

        This callback invokes the "per command" callbacks.
        
        cmd is the MEL command name and args is a list of strings
        that are the arguments of the command. The arguments have
        been split into their individual tokens. Quotes around
        quoted tokens are still present.
        
        Example:
        
        The MEL command setAttr -k off ".v"; would be passed in
        as onCommand('setAttr', ['-k', 'off', '"v"'])
        """
#        print "**",cmd, args
        # setAttr
        if cmd=="setAttr":
            args, opts = self.getOpt(args,
                                     self.setAttr_opt_def,
                                     self.setAttr_name_dict)
            self.onSetAttr(args[0], args[1:], opts)
        # createNode
        elif cmd=="createNode":
            args, opts = self.getOpt(args,
                                     self.createNode_opt_def,
                                     self.createNode_name_dict)
            self.onCreateNode(args[0], opts)
        # connectAttr
        elif cmd=="connectAttr":
            args, opts = self.getOpt(args,
                                     self.connectAttr_opt_def,
                                     self.connectAttr_name_dict)
            self.onConnectAttr(args[0], args[1], opts)
        # disconnectAttr
        elif cmd=="disconnectAttr":
            args, opts = self.getOpt(args,
                                     self.disconnectAttr_opt_def,
                                     self.disconnectAttr_name_dict)
            self.onDisconnectAttr(args[0], args[1], opts)
        # addAttr
        elif cmd=="addAttr":
            args, opts = self.getOpt(args,
                                     self.addAttr_opt_def,
                                     self.addAttr_name_dict)
            self.onAddAttr(opts)
        # parent
        elif cmd=="parent":
            args, opts = self.getOpt(args,
                                     self.parent_opt_def,
                                     self.parent_name_dict)
            self.onParent(args[:-1], args[-1], opts)
        # select
        elif cmd=="select":
            args, opts = self.getOpt(args,
                                     self.select_opt_def,
                                     self.select_name_dict)
            self.onSelect(args, opts)
        # fileInfo
        elif cmd=="fileInfo":
            args, opts = self.getOpt(args,
                                     self.fileInfo_opt_def,
                                     self.fileInfo_name_dict)
            self.onFileInfo(args[0], args[1], opts)
        # currentUnit
        elif cmd=="currentUnit":
            args, opts = self.getOpt(args,
                                     self.currentUnit_opt_def,
                                     self.currentUnit_name_dict)
            self.onCurrentUnit(opts)
        # requires
        elif cmd=="requires":
            args, opts = self.getOpt(args, {}, {})
            self.onRequires(args[0], args[1])
        # file
        elif cmd=="file":
            args, opts = self.getOpt(args,
                                     self.file_opt_def,
                                     self.file_name_dict)
            self.onFile(args[0], opts)
        # lockNode
        elif cmd=="lockNode":
            args, opts = self.getOpt(args,
                                     self.lockNode_opt_def,
                                     self.lockNode_name_dict)
            self.onLockNode(args, opts)
        # relationship
        elif cmd=="relationship":
            args, opts = self.getOpt(args,
                                     self.relationship_opt_def,
                                     self.relationship_name_dict)
            self.onRelationship(args, opts)
        # unknown
        else:
            sys.stderr.write("WARNING: %s, line %d: Unknown MEL command: '%s'\n"%(self.filename, self.cmd_start_linenr, cmd))


    # getOpt
    def getOpt(self, arglist, opt_def, name_dict):
        """Separate arguments from options and preprocess options.

        arglist is a list of arguments (i.e. the 'command line').
        opt_def specifies the available options and their respective
        number of arguments. name_dict is a dictionary that is used
        to convert short names into long names.

        The return value is a 2-tuple (args, opts) where args is a list
        of arguments and opts is a dictionary containing the options.
        The key is the long name of the option (without leading dash)
        and the value is a list of values. Any existing quotes around
        a value is removed (only around the option values, not around
        the args!).
        """

        args = []
        opts = {}
        
        i=0
        while i<len(arglist):
            arg = arglist[i]
            i += 1
            try:
                float(arg)
                is_number = True
            except:
                is_number = False
            # Option?
            a = stripQuotes(arg)
            if a[0:1]=="-" and not is_number:
                # Convert short names into long names...
                optname = name_dict.get(a[1:], a[1:])
                # Check if the option is known
                if optname not in opt_def:
                    raise SyntaxError("Unknown option in line %d: %s"%(self.cmd_start_linenr, optname))
                # Get the number of arguments
                numargs, filter = opt_def[optname]
                optvals = [stripQuotes(x) for x in arglist[i:i+numargs]]
                # Did the same option already appear? So this is a multi-use flag.
                # Then extend the current list with the new values
                if optname in opts:
                    opts[optname].extend(optvals)
                else:
                    opts[optname] = optvals
                i += numargs
            else:
                args.append(arg)

        return args, opts

    # processCommands
    def processCommands(self, s):
        """Process one or more commands.

        s is a string that contains one line of MEL code (may be several
        commands or only a partial command that is continued in the next
        line). This method splits the arguments and calls onCommand()
        for every command found.
        """
        # Split the command into tokens...
        a,n = self.splitCommand(s)
        if a!=[]:
            # Does a new command begin? then set the command name
            # and the args, otherwise just append to the existing args
            if self.new_cmd:
                self.cmd = a[0]
                self.args = a[1:]
                # Store the line number where the command began
                self.cmd_start_linenr = self.cpp.context.start_linenr
            else:
                self.args += a

        if n==-1:
            # The command isn't finished yet
            if self.cmd!=None:
                self.new_cmd = False
        else:
            # The command is finished, so execute it
            if self.cmd!=None:
                # Store the line number where the command ended
                self.cmd_end_linenr = self.cpp.context.linenr
                self.onCommand(self.cmd, self.args)
            self.new_cmd = True
            self.cmd = None
            self.args = []
            self.processCommands(s[n+1:])

    # splitCommand
    def splitCommand(self, s):
        """Split a command into its arguments.

        This is an extended version of the string split() method. It
        splits (using whitespace as separator) but takes quoted strings
        and ';' into account.
        Returns a list of strings and the position of the ';'
        that terminated the first command (or -1).
        The quotes around strings are kept.

        'setAttr -k off ".v";' -> (['setAttr', '-k', 'off', '".v"'], 19)
        """
        # Search for a quoted string
        b,e = self.findString(s)
        # Search for the first semicolon (which might be the true command
        # separator or not)
        n = s.find(";")
        # Was a semicolon before the first string? Then the string belongs
        # to a subsequent command, so ignore it for now
        if n!=-1 and (b is None or n<b):
            b = e = None
        # No string found?
        if b==None:
            if n==-1:
                return s.split(), -1
            else:
                return s[:n].split(), n
        else:
            if e==None:
                return s[:b].split() + [s[b:]+'"'], -1
            else:
                s2,n = self.splitCommand(s[e+1:])
                if n!=-1:
                    n += e+1
                return s[:b].split() + [s[b:e+1]] + s2, n

    # findString
    def findString(self, s):
        """Find the first string occurence.

        The return value is a 2-tuple (begin, end) with the indices
        of the opening and closing apostrophes (can also be None).

        'foo'             -> (None, None)
        'a="foo"'         -> (2, 6)
        'a="foo'          -> (2, None)
        'a="foo \" spam"' -> (2,14)
        """
        #'
        offset = 0
        while 1:
            # Search the beginning of a string
            n1 = s.find('"', offset)
            if n1==-1:
                return None,None
            # Search the end of the string (ignore quoted apostrophes)...
            start = n1+1
            n2 = None
            while 1:
                n2 = s.find('"', start)
                if n2==-1:
                    return n1,None
                elif s[n2-1]!='\\':
                    return n1,n2
                else:
                    start = n2+1

# DefaultMAReader
class DefaultMAReader(MAReader):
    """Default MA reader implementation.

    This class creates Node objects, sets attributes and does the
    connections so that after the file is read the entire dependency
    graph is available.

    A derived class only has to implement the end() callback and
    process the graph as desired. All created Node objects are available
    in the attribute self.nodelist.
    """

    def read(self, f):
        # A dict with imported Node objects
        # Key: Node name (without path) / Value: Node object
        # If the node name is not unique anymore, the value contains None.
        self.nodes = {}
        
        # A list with all Node objects (in the same order as they were
        # encountered in the file)
        self.nodelist = []
        
        # The currently active Node object
        # (changes with every createNode or select command)
        self.currentnode = None

        MAReader.read(self, f)

    # onCreateNode
    def onCreateNode(self, nodetype, opts):
        """Create a new node and make it current.
        """
        # Remove all quotes...
        nodetype = stripQuotes(nodetype)
#        for name in opts:
#            opts[name] = map(lambda x: stripQuotes(x), opts[name])
            
        node = self.createNode(nodetype, opts)
        self.currentnode = node

    # onSelect
    def onSelect(self, objects, opts):
        """Make another node current."""
        # Remove all quotes...
        objects = [stripQuotes(x) for x in objects]
        
        if opts!={"noExpand":[]}:
            raise ValueError("%s, %d: The select command contains unsupported options."%(self.filename, self.linenr))

        if len(objects)==0:
            raise ValueError("%s, %d: The select command contains no object name."%(self.filename, self.linenr))
        if len(objects)!=1:
            raise ValueError("%s, %d: The select command contains more than one object."%(self.filename, self.linenr))

        self.currentnode = self.findNode(objects[0], create=True)

    # onSetAttr
    def onSetAttr(self, attr, vals, opts):
        """Set an attribute."""
        if self.currentnode==None:
            return

        # Remove the quotes...
        attr = stripQuotes(attr)
#        for name in opts:
#            opts[name] = map(lambda x: stripQuotes(x), opts[name])

        if attr[0]!=".":
            sys.stderr.write("mayaascii: Warning: DefaultMAReader.onSetAttr(): The attribute refers to a different object than the current object. This is not yet supported.\n")

        self.currentnode.setAttr(attr, vals, opts)

    # onAddAttr
    def onAddAttr(self, opts):
        """Add a dynamic attribute."""
        if self.currentnode==None:
            return

        # Remove the quotes...
#        for name in opts:
#            opts[name] = map(lambda x: stripQuotes(x), opts[name])

        self.currentnode.addAttr(opts)

    # onConnectAttr
    def onConnectAttr(self, srcattr, dstattr, opts):
        """Make a connection.
        """

        # Remove the quotes...
        srcattr = stripQuotes(srcattr)
        dstattr = stripQuotes(dstattr)
#        for name in opts:
#            opts[name] = map(lambda x: stripQuotes(x), opts[name])

        # Split into object name and attribute name
        a = srcattr.split(".")
        snode = a[0]
        sattr = a[1]
        b = dstattr.split(".")
        dnode = b[0]
        dattr = b[1]
        
        sn = self.findNode(snode, create=True)
        dn = self.findNode(dnode, create=True)
        if sn!=None:
            if dn==None:
                sys.stderr.write('WARNING: %s, %d: connectAttr "%s" "%s"\n'%(self.filename, self.linenr, srcattr, dstattr))
                sys.stderr.write(' Node "%s" not found. The connection is ignored.\n'%dnode)
            else:
                sn.addOutConnection(sattr, dn, dnode, dattr)
        if dn!=None:
            dn.addInConnection(dattr, snode, sattr)


    # findNode
    def findNode(self, path, create=False):
        """Return the Node object corresponding to a particular path.

        path may also be None in which case None is returned.
        If create is True, any missing nodes are automatically created.

        (this method doesn't handle namespaces yet)
        """
        if path==None:
            return None
        
        namespace,names = splitDAGPath(path)
        # The current node (and eventually the result)
        node = None
        # Iterate over all names from 'top' to 'bottom'...
        for name in names:
            # An empty name? Then start from the beginning
            if name=="":
                node = None
            else:
                if node==None:
                    node = self.nodes.get(name)
                    if node==None:
                        if create:
                            node = self.createNode("<unknown>", {"name":[name]})
                        else:
                            raise KeyError("Node %s not found (%s is missing)"%(path, name))
                else:
                    for cn in node.iterChildren():
                        if name==cn.getName():
                            node = cn
                            break
                    else:
                        if create:
                            node = self.createNode("<unknown>", {"name":[name], "parent":[node.getFullName()]})
                        else:
                            raise KeyError("Node %s not found (%s is missing)"%(path, name))
        return node

    # createNode
    def createNode(self, nodetype, opts):
        """Create a new node and return it.
        """
        parentname = opts.get("parent", [None])[0]
        parent = self.findNode(parentname)
        node = Node(nodetype, opts, parent=parent)
        # The constant default name (if there was no name set in the file)
        # will override a previous node without name. But this doesn't
        # matter as the node cannot be addressed by name in the file anyway.
        nodename = node.getName()
        self.nodes[nodename] = node
        self.nodelist.append(node)
        return node
        


######################################################################

if __name__=="__main__":

    class TestReader(DefaultMAReader):

        def end(self):
            for node in self.nodelist:
                print(('%-30s %-20s parent:%s'%('"'+node.getFullName()+'"', node.nodetype, node.getParentName())))
                for attrname in node._setattr:
                    try:
                        val = node.getAttrValue(attrname, attrname, None, None)
                    except:
                        val = "<error retrieving value, need more type information>"
                    val = str(val)
                    if len(val)>60:
                        val = val[:60]+"..."
                    print(("  %s = %s"%(attrname, val)))
                


    maname = sys.argv[1]
    rd = TestReader()
    rd.read(maname)



from .externlib import *


# #### IMPORT EXTERNAL MODULES
# import bisect
# from collections import defaultdict
# from collections import Counter
# import copy
# import datetime as dt
# import fnmatch
# from ftplib import FTP
# import glob
# import inspect
# import itertools
# import linecache
# import math
# import matplotlib
# import matplotlib.pyplot as plt
# import multiprocessing as mp
# from natsort import natsorted, ns
# import numpy as np
# import numpy as npaa
# import operator
# import os 
# from os import remove, close
# import pandas as df
# import pandas as pd
# import pickle
# import scipy
# from scipy.signal import butter, lfilter, freqz
# import shutil
# import string
# import struct
# import sys
# import subprocess
# import re
# import tempfile
# from tempfile import mkstemp
# import tabulate
# import time
# import urllib
# import uuid




from .conv_coords            import *
from .conv_angle             import *
from .conv_geometric         import *
from .conv_proj_carto        import *
from .conv_rotation_matrices import *
from .conv_time              import *  



# ***** BEGIN LICENSE BLOCK *****
# Version: MPL 1.1/GPL 2.0/LGPL 2.1
#
# The contents of this file are subject to the Mozilla Public License Version
# 1.1 (the "License"); you may not use this file except in compliance with
# the License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/
#
# Software distributed under the License is distributed on an "AS IS" basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
# for the specific language governing rights and limitations under the
# License.
#
# The Original Code is the Python Computer Graphics Kit.
#
# The Initial Developer of the Original Code is Matthias Baas.
# Portions created by the Initial Developer are Copyright (C) 2004
# the Initial Developer. All Rights Reserved.
#
# Contributor(s):
#
# Alternatively, the contents of this file may be used under the terms of
# either the GNU General Public License Version 2 or later (the "GPL"), or
# the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
# in which case the provisions of the GPL or the LGPL are applicable instead
# of those above. If you wish to allow use of your version of this file only
# under the terms of either the GPL or the LGPL, and not to allow others to
# use your version of this file under the terms of the MPL, indicate your
# decision by deleting the provisions above and replace them with the notice
# and other provisions required by the GPL or the LGPL. If you do not delete
# the provisions above, a recipient may use your version of this file under
# the terms of any one of the MPL, the GPL or the LGPL.
#
# ***** END LICENSE BLOCK *****
# $Id: cgtypes.py,v 1.2 2006/02/14 19:29:39 mbaas Exp $

## \file cgtypes.py
## Contains the wrapped types vec3, vec4, mat3, mat4 and quat.

"""This module contains the basic types for doing 3D computations.

Classes:

- vec3: 3D vector (x,y,z)
- vec4: 4D vector (x,y,z,w) or (x,y,z,t)
- mat3: 3x3 matrix
- mat4: 4x4 matrix
- quat: Quaternion (w,x,y,z)

Functions:

- slerp(): Spherical linear interpolation between two quaternions.
- squad(): Spherical cubic interpolation.
- getEpsilon(): Return the threshold used for float comparisons
- setEpsilon(): Set a new threshold used for float comparisons
"""

from . import _core
from ._core import slerp, squad

# vec3
class vec3(_core.vec3):
    """A 3-dimensional vector of floats.

    This class is derived from the vec3 class in the _core module and
    adds some missing stuff that's better done in Python.
    """
    
    def __init__(self, *args):
        
        if len(args)==1:
            T = type(args[0])
            # scalar
            if T in [float, int]:
                _core.vec3.__init__(self, float(args[0]))
            # vec3  
            elif isinstance(args[0], vec3) or isinstance(args[0], _core.vec3):
                _core.vec3.__init__(self, *args)
            # Tuple/List
#            elif T in [tuple, list]:
#                _core.vec3.__init__(self, *args[0])
#                if len(args[0])==0:
#                    self.x = self.y = self.z = 0.0
#                elif len(args[0])==1:
#                    self.x = self.y = self.z = args[0][0]
#                elif len(args[0])==2:
#                    self.x, self.y = args[0]
#                    self.z         = 0.0
#                elif len(args[0])==3:
#                    self.x, self.y, self.z = args[0]
#                else:
#                    raise TypeError("vec3() takes at most 3 arguments")
            # String
            elif T is str:
                s=args[0].replace(","," ").replace("  "," ").strip().split(" ")
                if s==[""]:
                    s=[]
                f = [float(x) for x in s]
                _core.vec3.__init__(self, *f)
            # error
            else:
                try:
                    # Try to treat the argument as a sequence of 3 floats...
                    lst = list(args[0])
                    _core.vec3.__init__(self, *lst)
                except:
                    raise TypeError("vec3() arg can't be converted to vec3")

        else:
            # everything else that is not just one argument...
            _core.vec3.__init__(self, *args)

# vec4
class vec4(_core.vec4):
    """A 4-dimensional vector of floats.

    This class is derived from the vec4 class in the _core module and
    adds some missing stuff that's better done in Python.
    """
    
    def __init__(self, *args):
        
        if len(args)==1:
            T = type(args[0])
            # scalar
            if T in [float, int]:
                _core.vec4.__init__(self, float(args[0]))
            # vec4
            elif isinstance(args[0], vec4) or isinstance(args[0], _core.vec4):
                _core.vec4.__init__(self, *args)
            # String
            elif T is str:
                s=args[0].replace(","," ").replace("  "," ").strip().split(" ")
                if s==[""]:
                    s=[]
                f = [float(x) for x in s]
                _core.vec4.__init__(self, *f)
            # error
            else:
                try:
                    # Try to treat the argument as a sequence of floats...
                    lst = list(args[0])
                    _core.vec4.__init__(self, *lst)
                except:
                    raise TypeError("vec4() arg can't be converted to vec4")

        else:
            # everything else that is not just one argument...
            _core.vec4.__init__(self, *args)

# mat3
class mat3(_core.mat3):
    """A 3x3 matrix of floats.

    This class is derived from the mat3 class in the _core module and
    adds some missing stuff that's better done in Python.
    """

    def __init__(self, *args):
        
        if len(args)==1:
            T = type(args[0])
            # scalar
            if T in [float, int]:
                _core.mat3.__init__(self, float(args[0]))
            # mat3
            elif isinstance(args[0], mat3) or isinstance(args[0], _core.mat3):
                _core.mat3.__init__(self, *args)
            # Tuple/List
            elif T in [tuple, list]:
                _core.mat3.__init__(self, mat3(*args[0]))
            # String
            elif T is str:
                s=args[0].replace(","," ").replace("  "," ").strip().split(" ")
                if s==[""]:
                    s=[]
                f = [float(x) for x in s]
                _core.mat3.__init__(self, *f)
            # error
            else:
                raise TypeError("mat3() arg can't be converted to mat3")

        elif len(args)==3:
            try:
                c1,c2,c3 = args
                a,d,g = c1
                b,e,h = c2
                c,f,i = c3
                _core.mat3.__init__(self, a,b,c,d,e,f,g,h,i)
            except:
                raise TypeError("mat3() arg can't be converted to mat3")
        else:
            # everything else that is not just one argument...
            _core.mat3.__init__(self, *args)


    def __getitem__(self, key):
        
        T = type(key)
        if T is int:
            return _core.mat3.__getitem__(self, key)
        elif T is tuple:
            if len(key)!=2:
                raise ValueError("index tuple must be a 2-tuple")
            i,j = key
            if i<0 or i>2 or j<0 or j>2:
                raise IndexError("index out of range")
            return self[j][i]
        else:
            raise TypeError("index must be integer or 2-tuple")

# mat4
class mat4(_core.mat4):
    """A 4x4 matrix of floats.

    This class is derived from the mat4 class in the _core module and
    adds some missing stuff that's better done in Python.
    """

    def __init__(self, *args):
        
        if len(args)==1:
            T = type(args[0])
            # scalar
            if T in [float, int]:
                _core.mat4.__init__(self, float(args[0]))
            # mat4
            elif isinstance(args[0], mat4) or isinstance(args[0], _core.mat4):
                _core.mat4.__init__(self, *args)
            # Tuple/List
            elif T in [tuple, list]:
                _core.mat4.__init__(self, mat4(*args[0]))
            # String
            elif T is str:
                s=args[0].replace(","," ").replace("  "," ").strip().split(" ")
                if s==[""]:
                    s=[]
                f = [float(x) for x in s]
                _core.mat4.__init__(self, mat4(*f))
            # error
            else:
                raise TypeError("mat4() arg can't be converted to mat4")

        elif len(args)==4:
            try:
                c1,c2,c3,c4 = args
                _core.mat4.__init__(self)
                self.setColumn(0,vec4(c1))
                self.setColumn(1,vec4(c2))
                self.setColumn(2,vec4(c3))
                self.setColumn(3,vec4(c4))
            except:
                raise TypeError("mat4() arg can't be converted to mat4")
        elif len(args)==16:
            _core.mat4.__init__(self)
            self.setRow(0, vec4(args[0:4]))
            self.setRow(1, vec4(args[4:8]))
            self.setRow(2, vec4(args[8:12]))
            self.setRow(3, vec4(args[12:16]))
        else:
            # everything else that is not just one argument...
            _core.mat4.__init__(self, *args)

    def __getitem__(self, key):
        
        T = type(key)
        if T is int:
            return _core.mat4.__getitem__(self, key)
        elif T is tuple:
            if len(key)!=2:
                raise ValueError("index tuple must be a 2-tuple")
            i,j = key
            if i<0 or i>3 or j<0 or j>3:
                raise IndexError("index out of range")
            return self[j][i]
        else:
            raise TypeError("index must be integer or 2-tuple")

# quat
class quat(_core.quat):
    """A quaternion class.

    This class is derived from the quat class in the _core module and
    adds some missing stuff that's better done in Python.
    """
    
    def __init__(self, *args):
        
        if len(args)==1:
            T = type(args[0])
            # scalar
            if T in [float, int]:
                _core.quat.__init__(self, float(args[0]))
            # quat
            elif isinstance(args[0], quat) or isinstance(args[0], _core.quat):
                _core.quat.__init__(self, *args)
            # mat3/mat4
            elif isinstance(args[0], mat3) or isinstance(args[0], _core.mat3) or isinstance(args[0], mat4) or isinstance(args[0], _core.mat4):
                _core.quat.__init__(self)
                self.fromMat(args[0])
            # String
            elif T is str:
                s=args[0].replace(","," ").replace("  "," ").strip().split(" ")
                if s==[""]:
                    s=[]
                f = [float(x) for x in s]
                _core.quat.__init__(self, *f)
            # error
            else:
                try:
                    # Try to treat the argument as a sequence of floats...
                    lst = list(args[0])
                    _core.quat.__init__(self, *lst)
                except:
                    raise TypeError("quat() arg can't be converted to quat")

        # 2 arguments (angle & axis)
        elif len(args)==2:
            _core.quat.__init__(self)
            angle, axis = args
            self.fromAngleAxis(angle,axis)

        else:
            # everything else that is not just one argument...
            _core.quat.__init__(self, *args)

    def __pow__(self, other, modulo=None):
        """Return self**q."""
        if modulo!=None:
            raise TypeError("unsupported operation")
        q = quat(other)
        return (q*self.log()).exp()


def getEpsilon():
    """Return the threshold used for float comparisons.
    """
    return _core._getEpsilon()

def setEpsilon(eps):
    """Set the threshold used for float comparisons.
    """
    return _core._setEpsilon(eps)
